##cobra.solvers.glpk_solver
#This script provides wrappers for libglpk-java 1.0.22 and pyglpk 0.3
from warnings import warn
from copy import deepcopy
###solver specific parameters
from .parameters import status_dict, variable_kind_dict, \
     sense_dict, parameter_mappings, parameter_defaults, \
     objective_senses, default_objective_sense

from ..core.Solution import Solution
from time import time
solver_name = 'glpk'
sense_dict = eval(sense_dict[solver_name])
#Functions that are different for java implementation of a solver

from os import name
if name != "java":
    raise Exception("jython only")

warn("cobra.solvers.glpk_solver isn't mature.  consider using gurobi or cplex")
from org.gnu.glpk import GLPK, GLPKConstants, glp_smcp, glp_iocp
variable_kind_dict = eval(variable_kind_dict['%s_%s'%(solver_name,
                                            __name)])
status_dict = eval(status_dict['%s_%s'%(solver_name,
                                            __name)])
objective_senses = objective_senses['%s_%s'%(solver_name,
                                            __name)]
parameter_mappings = parameter_mappings['%s_%s'%(solver_name,
                                                 __name)]
parameter_defaults = parameter_defaults['%s_%s'%(solver_name,
                                                 __name)]

class Problem():
    """Create a more pythonesqe class to wrap the key
    features of the libglpk-java functions.
    
    """
    def __init__(self):
        """the attributes g, lp, mip should be made private
        """
        self._g = GLPK
        self._lp= GLPK.glp_create_prob()
        self._simplex_parameters = glp_smcp()
        self._mip_parameters = None
        self._g.glp_init_smcp(self._simplex_parameters)
        self.status = self.objective_value = None
        self._mip = False
    def set_name(self, name=''):
        self._g.glp_set_prob_name(self._lp, name)

    def solve(self):
        try:
            self._g.glp_simplex(self._lp,
                               self._simplex_parameters)
            if self._mip:
                #perform the MIP
                setattr(self._mip_parameters, 'msg_lev',
                         self._simplex_parameters.msg_lev)
                self._g.glp_intopt(self._lp, self._mip_parameters)
            self.status = self.get_status()
            self.objective_value = self.get_objective_value()
        except:
            self.status = 'failed'
        return self.status

    def get_status(self):
        if self._mip:
            status = self._g.glp_mip_status(self._lp)
        else:
            status = self._g.glp_get_status(self._lp)
        return status_dict[status]
    
    def set_objective_sense(self, parameter_value='maximize'):
        self._g.glp_set_obj_dir(self._lp,
                               eval(objective_senses[parameter_value]))

    def set_parameter(self, parameter_name, parameter_value, warning=False):
        if parameter_name == 'objective_sense':
            self.set_objective_sense(parameter_value)
        else:
            if parameter_name == 'meth' and parameter_value not in [1,2,3]:
                parameter_value = 1
            try:
                setattr(self._simplex_parameters, parameter_name,
                        parameter_value)
            except Exception, e1:
                try:
                    setattr(self._mip_parameters, parameter_name,
                            parameter_value)
                except Exception, e2:
                    if warning:
                        print "Could not set simplex parameter " +\
                              "%s: %s"%(parameter_name, repr(e1))
                        
                        if self._mip_parameters is not None:
                            print "Could not set mip parameter " +\
                                  "%s: %s"%(parameter_name, repr(e2))
    def get_objective_value(self):
        if self._mip:
            tmp_value = self._g.glp_mip_obj_val(self._lp)
        else:
            tmp_value = self._g.glp_get_obj_val(self._lp)
        return tmp_value

    def create_problem(self, cobra_model):
        g = self._g
        lp = self._lp
        number_of_reactions = len(cobra_model.reactions)
        number_of_metabolites = len(cobra_model.metabolites)
        g.glp_add_cols(lp, number_of_reactions)
        reaction_to_index = {}
        objective_dict = {}
        #Add in the variables
        tmp_kinds = []
        for i, the_reaction in enumerate(cobra_model.reactions):
            i_offset = i + 1
            reaction_to_index[the_reaction] = i_offset
            if the_reaction.objective_coefficient != 0:
                objective_dict[i_offset] = the_reaction.objective_coefficient
            g.glp_set_col_name(lp, i_offset, the_reaction.id)
            tmp_kinds.append(the_reaction.variable_kind)
            the_kind = variable_kind_dict[the_reaction.variable_kind]
            lower_bound = the_reaction.lower_bound
            upper_bound = the_reaction.upper_bound
            #Note. It is possible to have unbounded or one-bound variables
            if lower_bound == upper_bound:
                bound_kind = GLPKConstants.GLP_FX
            else:
                bound_kind = GLPKConstants.GLP_DB
            g.glp_set_col_kind(lp, i_offset, the_kind)
            g.glp_set_col_bnds(lp, i_offset,
                               bound_kind, the_reaction.lower_bound,
                               the_reaction.upper_bound)
        tmp_kinds = set(tmp_kinds)
        if 'integer' in tmp_kinds or 'binary' in tmp_kinds:
            self._mip = True
            self._mip_parameters = glp_iocp()
            g.glp_init_iocp(self._mip_parameters)
        #create constraints
        g.glp_add_rows(lp, number_of_metabolites)
        row_indices = []
        column_indices = []
        constraint_values = []
        for i, the_metabolite in enumerate(cobra_model.metabolites):
            i_offset = i + 1
            g.glp_set_row_name(lp, i_offset, the_metabolite.id)

            lower_bound = upper_bound = the_metabolite._bound
            constraint_sense = sense_dict[the_metabolite._constraint_sense]
            if constraint_sense == 'E':
                bound_type = GLPKConstants.GLP_FX
            elif constraint_sense == 'L':
                bound_type = GLPKConstants.GLP_UP
            elif constraint_sense == 'G':
                bound_type = GLPKConstants.GLP_LO
            elif constraint_sense == 'U':
                bound_type = GLPKConstants.GLP_FR
            elif hasattr(lower_bound, '__iter__'):
                lower_bound, upper_bound = lower_bound[:2]
                bound_type = GLPKConstants.GLP_DB
                

            g.glp_set_row_bnds(lp, i_offset, bound_type,
                               lower_bound, upper_bound)

            [(row_indices.append(i_offset),
              column_indices.append(reaction_to_index[k]),
              constraint_values.append(k._metabolites[the_metabolite]))
             for k in the_metabolite._reaction]

        #Load the constraints into the lp.  Need to use
        #typed arrays.
        number_of_constraints = len(row_indices)
        i_array = g.new_intArray(number_of_constraints)
        j_array = g.new_intArray(number_of_constraints)
        v_array = g.new_doubleArray(number_of_constraints)
        for a, (i, j, v) in enumerate(zip(row_indices,
                                          column_indices,
                                          constraint_values)):
            g.intArray_setitem(i_array, a+1, i)
            g.intArray_setitem(j_array, a+1, j)
            g.doubleArray_setitem(v_array, a+1, v)
        g.glp_load_matrix(lp, number_of_constraints, i_array,
                          j_array, v_array)
        # the following lines often cause memory crashes
        g.delete_intArray(i_array)
        g.delete_intArray(j_array)
        g.delete_doubleArray(v_array)
        

        g.glp_set_obj_name(lp, "z")
        [g.glp_set_obj_coef(lp, k, v)
          for k, v in objective_dict.iteritems()]

        




__solver_class = Problem

def set_parameter(lp, parameter_name, parameter_value):
    lp.set_parameter(parameter_name, parameter_value)


def get_status(lp):
    return lp.get_status()

def format_solution(lp, cobra_model, **kwargs):
    """

    """
    status = get_status(lp)
    if not lp._mip:
        try:
            x = [lp._g.glp_get_col_prim(lp._lp, i + 1)
                 for i in range(len(cobra_model.reactions))]
            x_dict = dict(zip(cobra_model.reactions, x))

            y = [lp._g.glp_get_row_dual(lp._lp, i + 1)
                 for i in range(len(cobra_model.metabolites))]
            y_dict = dict(zip(cobra_model.metabolites, y))
        
            objective_value = lp.objective_value
        except Exception, e:
            print repr(e)
            y = y_dict = x = x_dict = objective_value = None
            #print status
    else:
        try:
            x = [lp._g.glp_mip_col_val(lp._lp, i + 1)
                 for i in range(len(cobra_model.reactions))]
            x_dict = dict(zip(cobra_model.reactions, x))
            y = y_dict = None
            objective_value = lp.objective_value
        except:
            y = y_dict = x = x_dict = objective_value = None

    return(Solution(objective_value, x=x, x_dict=x_dict, y=y,
                    y_dict=y_dict, status=status))
def create_problem(cobra_model,  **kwargs):
    """Solver-specific method for constructing a solver problem from
    a cobra.Model.  This can be tuned for performance using kwargs


    """
    the_parameters = parameter_defaults
    if kwargs:
        the_parameters = deepcopy(parameter_defaults)
        the_parameters.update(kwargs)
    quadratic_component = the_parameters['quadratic_component']
    new_objective = the_parameters['new_objective']
    if quadratic_component is not None:
        raise Exception('%s cannot solve QPs, try a different solver'%solver_name)
    lp = Problem()        # Create empty problem instance
    lp.create_problem(cobra_model)
    [set_parameter(lp, parameter_mappings[k], v)
     for k, v in the_parameters.iteritems() if k in parameter_mappings]
    return(lp)

def update_problem(lp, cobra_model, **kwargs):
    """
    Assumes that neither Metabolites nor Reaction have been
    added or removed.

    Currently only deals with reaction bounds and objective
    coefficients.

    """
    g = lp._g
    l = lp._lp
    for i, the_reaction in enumerate(cobra_model.reactions):
        lower_bound = float(the_reaction.lower_bound)
        upper_bound = float(the_reaction.upper_bound)
        objective_coefficient = float(the_reaction.objective_coefficient)
        if lower_bound == upper_bound:
            bound_type = GLPKConstants.GLP_FX
        else:
            bound_type = GLPKConstants.GLP_DB
        g.glp_set_col_bnds(l, i + 1, bound_type, lower_bound, upper_bound)
        g.glp_set_obj_coef(l, i + 1, objective_coefficient)

    
def solve_problem(lp, **kwargs):
    """A performance tunable method for updating a model problem file

    """
    #Update parameter settings if provided
    if kwargs:
        [set_parameter(lp, parameter_mappings[k], v)
         for k, v in kwargs.iteritems() if k in parameter_mappings]
    try:
        print_solver_time = kwargs['print_solver_time']
        start_time = time()
    except:
        print_solver_time = False
    lp_method = lp._simplex_parameters.meth
    lp.solve()
    status = get_status(lp)
    if print_solver_time:
        print 'optimize time: %f'%(time() - start_time)
    return status

    
def solve(cobra_model, **kwargs):
    """Smart interface to optimization solver functions that will convert
    the cobra_model to a solver object, set the parameters, and try multiple
    methods to get an optimal solution before returning the solver object and
    a cobra.Solution (which is attached to cobra_model.solution)

    cobra_model: a cobra.Model

    returns a dict: {'the_problem': solver specific object, 'the_solution':
    cobra.Solution for the optimization problem'}
    

    """
    #Start out with default parameters and then modify if
    #new onese are provided
    the_parameters = deepcopy(parameter_defaults)
    if kwargs:
        the_parameters.update(kwargs)
    #Update objectives if they are new.
    error_reporting = the_parameters['error_reporting']
    if 'new_objective' in the_parameters and \
           the_parameters['new_objective'] not in ['update problem', None]:
       from ..flux_analysis.objective import update_objective
       update_objective(cobra_model, the_parameters['new_objective'])
    if 'the_problem' in the_parameters:
        the_problem = the_parameters['the_problem']
    else:
        the_problem = None
    if isinstance(the_problem, __solver_class):
        #Update the problem with the current cobra_model
        lp = the_problem
        update_problem(lp, cobra_model, **the_parameters)
    else:
        #Create a new problem
        lp = create_problem(cobra_model, **the_parameters)
    #Deprecated way for returning a solver problem created from a cobra_model
    #without performing optimization
    if the_problem == 'setup':
        return lp
    ###Try to solve the problem using other methods if the first method doesn't work
    lp_method = the_parameters['lp_method']
    the_methods = [1, 2, 3]
    if lp_method in the_methods:
        the_methods.remove(lp_method)
    #Start with the user specified method
    the_methods.insert(0, lp_method)
    for the_method in the_methods:
        the_parameters['lp_method'] = the_method
        try:
            status = solve_problem(lp, **the_parameters)
        except:
            status = 'failed'
        if status == 'optimal':
            break
    
    the_solution = format_solution(lp, cobra_model)
    if status != 'optimal' and error_reporting:
        print '%s failed: %s'%(solver_name, status)
    cobra_model.solution = the_solution
    solution = {'the_problem': lp, 'the_solution': the_solution}
    return solution



from time import time
from warnings import warn
from copy import deepcopy

from ..external.six import string_types, iteritems

from ..manipulation import delete_model_genes, undelete_model_genes
from ..manipulation.delete import find_gene_knockout_reactions
from ..solvers import solver_dict, get_solver_name


nan = float('nan')


try:
    from .moma import moma
except Exception as e:
    def moma(**kwargs):
        warn("moma is currently not functional")


def single_deletion(cobra_model, element_list=None,
                    method='fba', element_type='gene', solver=None):
    """Wrapper for single_gene_deletion and the single_reaction_deletion
    functions

    
    cobra_model: a cobra.Model object

    element_list: Is None or a list of elements (genes or reactions) to
    delete.

    method: 'fba' or 'moma'

    the_problem: Is None, 'return', or an LP model object for the solver.

    element_type: 'gene' or 'reaction'

    solver: 'glpk', 'gurobi', or 'cplex'.

    error_reporting: None or True to disable or enable printing errors encountered
    when trying to find the optimal solution.

    discard_problems: Boolean.  If True do not save problems.  This will
    help with memory issues related to gurobi.
    .. warning:: This is deprecated.

    Returns a list of dictionaries: growth_rate_dict, solution_status_dict,
    problem_dict where the key corresponds to each element in element_list.

    """
    if solver is None:
        solver = get_solver_name() if method == "fba" else get_solver_name(qp=True)
    # fast versions of functions
    if method == "fba":
        if element_type == "gene":
            return single_gene_deletion_fba(cobra_model, element_list,
                                             solver=solver)
        elif element_type == "reaction":
            return single_reaction_deletion_fba(cobra_model, element_list,
                                                 solver=solver)
    if element_type == 'gene':
        the_solution = single_gene_deletion(cobra_model, element_list,
                                    method=method, solver=solver)

    else:
        the_solution = single_reaction_deletion(cobra_model, element_list,
                                        method=method, solver=solver)
    return the_solution


def single_reaction_deletion_fba(cobra_model, reaction_list=None, solver=None):
    solver = solver_dict[get_solver_name() if solver is None else solver]
    lp = solver.create_problem(cobra_model)
    growth_rate_dict = {}
    status_dict = {}
    if reaction_list is None:
        reaction_list = cobra_model.reactions
    else:
        reaction_list = [cobra_model.reactions.get_by_id(i) \
                         if isinstance(i, string_types) else i \
                         for i in reaction_list]
    for reaction in reaction_list:
        old_bounds = (reaction.lower_bound, reaction.upper_bound)
        index = cobra_model.reactions.index(reaction)
        solver.change_variable_bounds(lp, index, 0., 0.)
        solver.solve_problem(lp)
        status = solver.get_status(lp)
        status_dict[reaction.id] = status
        growth_rate_dict[reaction.id] = solver.get_objective_value(lp) if status == "optimal" else 0.
        # reset the problem
        solver.change_variable_bounds(lp, index, old_bounds[0], old_bounds[1])
    return(growth_rate_dict, status_dict)

def single_gene_deletion_fba(cobra_model, gene_list=None, solver=None):
    solver = solver_dict[get_solver_name() if solver is None else solver]
    lp = solver.create_problem(cobra_model)
    growth_rate_dict = {}
    status_dict = {}
    if gene_list is None:
        gene_list = cobra_model.genes
    else:
        gene_list = [cobra_model.genes.get_by_id(i) \
                     if isinstance(i, string_types) else i for i in gene_list]
    for gene in gene_list:
        old_bounds = {}
        for reaction in find_gene_knockout_reactions(cobra_model, [gene]):
            index = cobra_model.reactions.index(reaction)
            old_bounds[index] = (reaction.lower_bound, reaction.upper_bound)
            solver.change_variable_bounds(lp, index, 0., 0.)
        solver.solve_problem(lp)
        status = solver.get_status(lp)
        status_dict[gene.id] = status
        growth_rate_dict[gene.id] = solver.get_objective_value(lp)# if status == "optimal" else 0.
        # reset the problem
        for index, bounds in iteritems(old_bounds):
            solver.change_variable_bounds(lp, index, bounds[0], bounds[1])
    return(growth_rate_dict, status_dict)

def single_reaction_deletion(cobra_model, element_list=None,
                             method='fba', the_problem='return',
                             solver=None, error_reporting=None,
                             discard_problems=True):
    """Performs optimization simulations to realize the objective defined
    from cobra_model.reactions[:].objective_coefficients after deleting each reaction
    from the model.
    
    cobra_model: a cobra.Model object

    element_list: Is None or a list of cobra.Reactions in cobra_model to disable.
    If None then disable each reaction in cobra_model.reactions and optimize for the
    objective function defined from cobra_model.reactions[:].objective_coefficients.

    method: 'fba' is the only option at the moment.

    the_problem: Is None, 'reuse', or an LP model object for the solver.

    solver: 'glpk', 'gurobi', or 'cplex'.

    discard_problems: Boolean.  If True do not save problems.  This will
    help with memory issues related to gurobi.
    
    Returns a list of dictionaries: growth_rate_dict, solution_status_dict,
    problem_dict where the key corresponds to each reaction in reaction_list.

    """
    if solver is None:
        solver = get_solver_name() if method == "fba" else get_solver_name(qp=True)
    #element_list so we can merge single_reaction_deletion and single_gene_deletion

    wt_model = cobra_model.copy() #Original wild-type (wt) model.
    wt_model.id = 'Wild-Type'
    #MOMA constructs combined quadratic models thus we cannot reuse a model
    #generated by the cobra_model.optimize call
    if method.lower() == 'moma':
        the_problem = 'return'
        mutant_model = wt_model.copy() #Need a second model for moma
    else:
        mutant_model = cobra_model
    discard_problems = False
    if the_problem:
        the_problem = 'return'
        discard_problems = True
    the_problem = wt_model.optimize(the_problem=the_problem, solver=solver,
                                       error_reporting=error_reporting)
    wt_f = wt_model.solution.f
    wt_status = wt_model.solution.status
    wt_x = deepcopy(wt_model.solution.x)
    wt_x_dict = deepcopy(wt_model.solution.x_dict)

    wt_problem = the_problem
    if element_list is None:
        element_list = mutant_model.reactions
    elif not hasattr(element_list[0], 'id'):
        element_list = map(mutant_model.reactions.get_by_id, element_list)
    else:
        if mutant_model is not cobra_model:
            element_list = [x.id for x in element_list]
            element_list = map(mutant_model.genes.get_by_id, element_list)

    growth_rate_dict = {}
    solution_status_dict = {}
    problem_dict = {}
    combined_model = None
    for the_element in element_list:
        #delete the gene
        #if the deletion alters the bounds then run simulation
        old_lower_bound, old_upper_bound = map(float, [the_element.lower_bound,
                                                       the_element.upper_bound])
        mutant_model.id = the_element.id
        if old_lower_bound != 0 or old_upper_bound != 0:
            the_element.lower_bound = 0
            the_element.upper_bound = 0
            if method.lower() == 'fba':
                the_problem = mutant_model.optimize(the_problem=wt_problem,
                                                    solver=solver,
                                                    error_reporting=error_reporting)
                growth_rate_dict[the_element] = mutant_model.solution.f
                solution_status_dict[the_element] = mutant_model.solution.status
            elif method.lower() == 'moma':
                try:
                    #TODO: Need to figure out why reusing the problem and the combined_model do not
                    #speed things up here.
                    moma_solution = moma(wt_model, mutant_model, solver=solver, the_problem=the_problem,
                                         combined_model=combined_model)
                    the_problem = moma_solution.pop('the_problem')
                    growth_rate_dict[the_element] = float(moma_solution.pop('objective_value'))
                    solution_status_dict[the_element] = moma_solution.pop('status')
                    combined_model = moma_solution.pop('combined_model')
                except:
                    growth_rate_dict[the_element] = nan
                    the_problem = None
                    solution_status_dict[the_element] = 'failed'

            if discard_problems:
                problem_dict[the_element] = 'discarded'
            else:
                problem_dict[the_element] = the_problem
            if not the_problem:
                the_problem = wt_problem
            #reset the model
            the_element.lower_bound = old_lower_bound
            the_element.upper_bound = old_upper_bound
        #else just use the wt_f and x
        else:
            if discard_problems:
                problem_dict[the_element] = 'discarded'
            else:
                problem_dict[the_element] = wt_problem
            growth_rate_dict[the_element] = wt_f
            solution_status_dict[the_element] = wt_status
    
    
    return(growth_rate_dict, solution_status_dict, problem_dict)

def single_gene_deletion(cobra_model, element_list=None,
                         method='fba', the_problem='reuse', solver=None,
                         error_reporting=None):
    """Performs optimization simulations to realize the objective defined
    from cobra_model.reactions[:].objective_coefficients after deleting each gene in
    gene_list from the model.
    
    cobra_model: a cobra.Model object

    element_list: Is None or a list of genes to delete.  If None then
    disable each reaction associated with each gene in cobra_model.genes.

    method: 'fba' or 'moma'

    the_problem: Is None or 'reuse'.

    solver: 'glpk', 'gurobi', or 'cplex'.

    Returns a list of dictionaries: growth_rate_dict, solution_status_dict,
    problem_dict where the key corresponds to each reaction in reaction_list.

    TODO: Add in a section that allows copying and collection of problem for
    debugging purposes.

    """
    if solver is None:
        solver = get_solver_name() if method == "fba" else get_solver_name(qp=True)
    wt_model = cobra_model.copy() #Original wild-type (wt) model.
    wt_model.id = 'Wild-Type'
    #MOMA constructs combined quadratic models thus we cannot reuse a model
    #generated by the cobra_model.optimize call
    if method.lower() == 'moma':
        the_problem = 'return'
        mutant_model = wt_model.copy() #Need a second model for moma
    else:
        mutant_model = cobra_model
    discard_problems = False
    if the_problem:
        the_problem = 'return'
        discard_problems = True

    solver_object = solver_dict[solver]
    the_problem = solver_object.create_problem(wt_model)
    solver_object.solve_problem(the_problem)
    solution = solver_object.format_solution(the_problem, wt_model)
    wt_f = solution.f
    wt_status = solution.status
    wt_x = deepcopy(solution.x)
    wt_x_dict = deepcopy(solution.x_dict)

    if element_list is None:
        element_list = mutant_model.genes
    elif not hasattr(element_list[0], 'id'):
        element_list = map(mutant_model.genes.get_by_id, element_list)
    else:
        if mutant_model is not cobra_model:
            element_list = [x.id for x in element_list]
            element_list = map(mutant_model.genes.get_by_id, element_list)
    wt_problem = the_problem

    growth_rate_dict = {}
    solution_status_dict = {}
    problem_dict = {}
    combined_model = None

    for the_element in element_list:
        #delete the gene
        #if the deletion alters the bounds then run simulation
        delete_model_genes(mutant_model, the_element)
        mutant_model.id = the_element.id
        if mutant_model._trimmed:
            if method.lower() == 'fba':
                the_problem = mutant_model.optimize(the_problem=wt_problem,
                                                    solver=solver,
                                                    error_reporting=error_reporting)
                growth_rate_dict[the_element.id] = mutant_model.solution.f
                solution_status_dict[the_element.id] = mutant_model.solution.status
            elif method.lower() == 'moma':
                try:
                    #TODO: Need to figure out why reusing the problem and the combined_model do not
                    #speed things up here.
                    moma_solution = moma(wt_model, mutant_model, solver=solver, the_problem=the_problem,
                                         combined_model=combined_model)
                    the_problem = moma_solution.pop('the_problem')
                    growth_rate_dict[the_element.id] = float(moma_solution.pop('objective_value'))
                    solution_status_dict[the_element.id] = moma_solution.pop('status')
                    combined_model = moma_solution.pop('combined_model')
                except:
                    growth_rate_dict[the_element.id] = nan
                    the_problem = None
                    solution_status_dict[the_element.id] = 'failed'

            if discard_problems:
                problem_dict[the_element.id] = 'discarded'
            else:
                problem_dict[the_element.id] = the_problem
            if not the_problem:
                the_problem = wt_problem
            #reset the model
            undelete_model_genes(mutant_model)
        #else just use the wt_f and x
        else:
            if discard_problems:
                problem_dict[the_element.id] = 'discarded'
            else:
                problem_dict[the_element.id] = wt_problem
            growth_rate_dict[the_element.id] = wt_f
            solution_status_dict[the_element.id] = wt_status
    
    
    return(growth_rate_dict, solution_status_dict, problem_dict)




import re

from numpy import array, object as np_object
from scipy.io import loadmat, savemat
from scipy.sparse import coo_matrix

from .. import Model, Metabolite, Reaction, Formula

# try to use an ordered dict
try:
    from scipy.version import short_version
    scipy_version = int(short_version.split(".")[1])
    # if scipy version is earlier than 0.11, OrderedDict will not work, so use
    # a dict instead
    if scipy_version < 11:
        dicttype = dict
    else:
        from collections import OrderedDict as dicttype
    del short_version, scipy_version
except ImportError:
    dicttype = dict


# precompiled regular expressions
_bracket_re = re.compile("r\[[a-z]\]$")
_underscore_re = re.compile(r"_[a-z]$")


def _get_id_comparment(id):
    """extract the compartment from the id string"""
    bracket_search = _bracket_re.findall(id)
    if len(bracket_search) == 1:
        return bracket_search[0][1]
    underscore_search = _underscore_re.findall(id)
    if len(underscore_search) == 1:
        return underscore_search[0][1]
    return None


def _cell(x):
    """translate an array x into a MATLAB cell array"""
    return array(x, dtype=np_object)


def load_matlab_model(infile_path, variable_name=None):
    """Load a cobra model stored as a .mat file

    infile_path : str

    variable_name : str, optional
        The variable name of the model in the .mat file. If this is not
        specified, then the first MATLAB variable which looks like a COBRA
        model will be used

    """
    data = loadmat(infile_path)
    if variable_name is not None:
        return from_mat_struct(data[variable_name], model_id=variable_name)
    else:
        # will try all of the variables in the dict
        possible_names = {}
        for key in data.keys():
            possible_names[key] = None
        # skip meta variables
        to_remove = ["__globals__", "__header__", "__version__"]
        to_pop = []
        for name in possible_names:
            if name in to_remove:
                to_pop.append(name)
        for i in to_pop:
            possible_names.pop(i)
        possible_names = possible_names.keys()
    for possible_name in possible_names:
        try:
            return from_mat_struct(data[possible_name], model_id=possible_name)
        except ValueError:
            None
    # If code here is executed, then no model was found.
    raise Exception("no COBRA model found")


def save_matlab_model(model, file_name):
    """Save the cobra model as a .mat file.

    This .mat file can be used directly in the MATLAB version of COBRA.

    model : :class:`~cobra.core.Model.Model` object

    file_name : str or file-like object

    """
    mat = create_mat_dict(model)
    savemat(file_name, {str(model.description): mat},
            appendmat=True, oned_as="column")


def create_mat_dict(model):
    """create a dict mapping model attributes to arrays"""
    model = model.to_array_based_model()
    rxns = model.reactions
    mets = model.metabolites
    mat = dicttype()
    mat["mets"] = _cell(mets.list_attr("id"))
    mat["metNames"] = _cell(mets.list_attr("name"))
    mat["metFormulas"] = _cell([str(m.formula) for m in mets])
    mat["genes"] = _cell(model.genes.list_attr("id"))
    mat["grRules"] = _cell(rxns.list_attr("gene_reaction_rule"))
    mat["rxns"] = _cell(rxns.list_attr("id"))
    mat["rxnNames"] = _cell(rxns.list_attr("name"))
    mat["subSystems"] = _cell(rxns.list_attr("subsystem"))
    mat["csense"] = "".join(model._constraint_sense)
    mat["S"] = model.S
    # multiply by 1 to convert to float, working around scipy bug
    # https://github.com/scipy/scipy/issues/4537
    mat["lb"] = array(rxns.list_attr("lower_bound")) * 1.
    mat["ub"] = array(rxns.list_attr("upper_bound")) * 1.
    mat["b"] = array(mets.list_attr("_bound")) * 1.
    mat["c"] = array(rxns.list_attr("objective_coefficient")) * 1.
    mat["rev"] = array(rxns.list_attr("reversibility"))
    mat["description"] = str(model.description)
    return mat


def from_mat_struct(mat_struct, model_id=None):
    """create a model from the COBRA toolbox struct

    The struct will be a dict read in by scipy.io.loadmat

    """
    m = mat_struct
    if m.dtype.names is None:
        raise ValueError("not a valid mat struct")
    if not set(["rxns", "mets", "S", "lb", "ub"]) <= set(m.dtype.names):
        raise ValueError("not a valid mat struct")
    model = Model()
    if "description" in m:
        model.id = m["description"][0, 0][0]
    elif model_id is not None:
        model.id = model_id
    else:
        model.id = "imported_model"
    model.description = model.id
    for i, name in enumerate(m["mets"][0, 0]):
        new_metabolite = Metabolite()
        new_metabolite.id = str(name[0][0])
        new_metabolite.compartment = _get_id_comparment(new_metabolite.id)
        try:
            new_metabolite.name = str(m["metNames"][0, 0][i][0][0])
        except (IndexError, ValueError):
            pass
        try:
            new_metabolite.formula = \
                Formula(str(m["metFormulas"][0][0][i][0][0]))
        except (IndexError, ValueError):
            pass
        model.add_metabolites([new_metabolite])
    new_reactions = []
    for i, name in enumerate(m["rxns"][0, 0]):
        new_reaction = Reaction()
        new_reaction.id = str(name[0][0])
        new_reaction.lower_bound = float(m["lb"][0, 0][i][0])
        new_reaction.upper_bound = float(m["ub"][0, 0][i][0])
        new_reaction.objective_coefficient = float(m["c"][0, 0][i][0])
        try:
            new_reaction.gene_reaction_rule = str(m['grRules'][0, 0][i][0][0])
        except (IndexError, ValueError):
            pass
        try:
            new_reaction.name = str(m["rxnNames"][0, 0][i][0][0])
        except (IndexError, ValueError):
            pass
        try:
            new_reaction.subsystem = str(m['subSystems'][0, 0][i][0][0])
        except (IndexError, ValueError):
            pass
        new_reactions.append(new_reaction)
    model.add_reactions(new_reactions)
    coo = coo_matrix(m["S"][0, 0])
    for i, j, v in zip(coo.row, coo.col, coo.data):
        model.reactions[j].add_metabolites({model.metabolites[i]: v})
    return model



import re
from warnings import warn

from .Species import Species


class Gene(Species):

    """A Gene is a special class of metabolite.


    TODO: Make design decisions about TUs and such
    """

    def __init__(self, id, formula=None,
                 name=None, compartment=None, strand='+',
                 locus_start=0, locus_end=0, functional=True):
        """
        id: A string.

        formula: cobra.Formula or a chemical formula str.  Defaults to None
        to save time in pickling and such.

        name: String.  A human readable name.

        compartment: None or a dictionary indicating the cellular location
        of the metabolite.  Used when in a cobra.Reaction or Model
        object

        strand: '+' or '-' to indicate forward or reverse strand for DNA.

        locus_start: Int.  The index of the starting base for the gene.

        locus_end: Int. The index of the last base for the gene.

        functional: Boolean.  Indicate whether the gene is functional.  If it
        is not functional then it cannot be used in an enzyme complex nor
        can its products be used.

        """
        Species.__init__(self, id, formula=formula,
                         name=name, compartment=compartment)
        self.locus_start = locus_start
        self.locus_end = locus_end
        self.strand = strand
        self.functional = functional

    def remove_from_model(self, model=None,
                          make_dependent_reactions_nonfunctional=True):
        """Removes the association

        make_dependent_reactions_nonfunctional: Boolean.  If True then replace
        the gene with 'False' in the gene association, else replace the gene
        with 'True'

        .. note :: Simulating gene knockouts is much better handled by
                   cobra.manipulation.delete_model_genes

        """
        if model is not None:
            warn("passing the model in is unnecessary and deprecated")
            if model != self._model:
                raise Exception("%s is a member of %s, not %s" %
                                (repr(self), repr(self._model), repr(model)))
        if self._model is None:
            raise Exception('%s is not in a model' % repr(self))

        if make_dependent_reactions_nonfunctional:
            gene_state = 'False'
        else:
            gene_state = 'True'
        the_gene_re = re.compile('(^|(?<=( |\()))%s(?=( |\)|$))' %
                                 re.escape(self.id))

        self._model.genes.remove(self)
        self._model = None

        for the_reaction in list(self._reaction):
            the_reaction._gene_reaction_rule = the_gene_re.sub(
                gene_state, the_reaction.gene_reaction_rule)
            the_reaction._genes.remove(self)
            # Now, deactivate the reaction if its gene association evaluates
            # to False
            the_gene_reaction_relation = the_reaction.gene_reaction_rule
            for other_gene in the_reaction._genes:
                other_gene_re = re.compile('(^|(?<=( |\()))%s(?=( |\)|$))' %
                                           re.escape(other_gene.id))
                the_gene_reaction_relation = other_gene_re.sub(
                    'True',
                    the_gene_reaction_relation)

            if not eval(the_gene_reaction_relation):
                the_reaction.lower_bound = 0
                the_reaction.upper_bound = 0
        self._reaction.clear()



from multiprocessing import Queue, Process, cpu_count

from ..solvers import get_solver_name, solver_dict
from ..external.six import iteritems


def compute_fba_deletion_worker(cobra_model, solver, job_queue, output_queue,
                                **kwargs):
    solver = solver_dict[get_solver_name() if solver is None else solver]
    lp = solver.create_problem(cobra_model)
    solver_args = kwargs
    solver.solve_problem(lp)
    while True:
        indexes, label = job_queue.get()
        label = indexes if label is None else label
        result = compute_fba_deletion(lp, solver, cobra_model, indexes,
                                      **solver_args)
        output_queue.put((label, result))


def compute_fba_deletion(lp, solver_object, model, indexes, **kwargs):
    s = solver_object
    old_bounds = {}
    for i in indexes:
        reaction = model.reactions[i]
        old_bounds[i] = (reaction.lower_bound, reaction.upper_bound)
        s.change_variable_bounds(lp, i, 0., 0.)
    try:
        s.solve_problem(lp, **kwargs)
    except Exception as e:
        return RuntimeError("solver failure when deleting %s: %s" %
                                (str(indexes), repr(e)))
    status = s.get_status(lp)
    objective = s.get_objective_value(lp) if status == "optimal" else 0.

    # reset the problem, which must be done after reading the solution
    for index, bounds in iteritems(old_bounds):
        s.change_variable_bounds(lp, index, bounds[0], bounds[1])

    if status == "infeasible" or status == "optimal":
        return objective
    else:
        return RuntimeError("solver failure (status %s) for when deleting %s" %
                            (status, str(indexes)))


class CobraDeletionPool(object):
    """A pool of workers for solving deletions

    submit jobs to the pool using submit and recieve results using receive_all
    """
    # Having an existing basis makes solving an existing LP much faster. The
    # most efficient approach is to have a worker function which modifies an LP
    # object and reverts it back after each calculation. Each lp object stores
    # the basis so subsequent LP's are solved more quickely, and memory does
    # not need to be re-allocated each time to create a new problem. Because
    # state is being saved, the workers in the deletion pool are careful about
    # reverting the object after simulating a deletion, and are written to be
    # flexible enough so they can be used in most applications instead of
    # writing a custom worker each time.
    def __init__(self, cobra_model, n_processes=None, solver=None, **kwargs):
        if n_processes is None:
            n_processes = min(cpu_count(), 4)
        # start queues
        self.job_queue = Queue()  # format is (indexes, job_label)
        self.n_submitted = 0
        self.n_complete = 0
        self.output_queue = Queue()  # format is (job_label, growth_rate)
        # start processes
        self.processes = []
        for i in range(n_processes):
            p = Process(target=compute_fba_deletion_worker,
                        args=[cobra_model, solver,
                              self.job_queue, self.output_queue],
                        kwargs=kwargs)
            self.processes.append(p)

    def start(self):
        for p in self.processes:
            p.start()

    def terminate(self):
        for p in self.processes:
            p.terminate()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.terminate()

    def submit(self, indexes, label=None):
        self.job_queue.put((indexes, label))
        self.n_submitted += 1

    def receive_one(self):
        """This function blocks"""
        self.n_complete += 1
        result = self.output_queue.get()
        if isinstance(result[1], Exception):
            raise result[1]
        return result

    def receive_all(self):
        while self.n_complete < self.n_submitted:
            self.n_complete += 1
            result = self.output_queue.get()
            if isinstance(result[1], Exception):
                raise result[1]
            yield result

    @property
    def pids(self):
        return [p.pid for p in self.processes]

    def __del__(self):
        for process in self.processes:
            process.terminate()
            process.join()


class CobraDeletionMockPool(object):
    """Mock pool solves LP's in the same process"""
    def __init__(self, cobra_model, n_processes=1, solver=None, **kwargs):
        if n_processes != 1:
            from warnings import warn
            warn("Mock Pool does not do multiprocessing")
        self.job_queue = []
        self.solver_args = kwargs
        solver_name = get_solver_name() if solver is None else solver
        self.solver = solver_dict[solver_name]
        self.lp = self.solver.create_problem(cobra_model)
        self.solver.solve_problem(self.lp)
        self.model = cobra_model

    def submit(self, indexes, label=None):
        self.job_queue.append((indexes, label))

    def receive_one(self):
        indexes, label = self.job_queue.pop()
        result = compute_fba_deletion(self.lp, self.solver, self.model,
                                      indexes, **self.solver_args)
        if isinstance(result, Exception):
            raise result
        return (label, result)

    def receive_all(self):
        for i in range(len(self.job_queue)):
            indexes, label = self.job_queue.pop()
            result = compute_fba_deletion(self.lp, self.solver, self.model,
                                          indexes, **self.solver_args)
            if isinstance(result, Exception):
                raise result
            yield (label, result)

    def start(self):
        None

    def terminate(self):
        None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        None



# Interface to gurobipy

from warnings import warn
from itertools import izip

from gurobipy import Model, LinExpr, GRB, QuadExpr


from ..core.Solution import Solution
from ..external.six import string_types

solver_name = 'gurobi'
_SUPPORTS_MILP = True


# set solver-specific parameters
parameter_defaults = {'objective_sense': 'maximize',
                      'tolerance_optimality': 1e-6,
                      'tolerance_feasibility': 1e-6,
                      'tolerance_integer': 1e-9,
                      # This is primal simplex, default is -1 (automatic)
                      'lp_method': 0,
                      'verbose': False,
                      'log_file': ''}
parameter_mappings = {'log_file': 'LogFile',
                      'lp_method': 'Method',
                      'threads': 'Threads',
                      'objective_sense': 'ModelSense',
                      'output_verbosity': 'OutputFlag',
                      'verbose': 'OutputFlag',
                      'quadratic_precision': 'Quad',
                      'time_limit': 'TimeLimit',
                      'tolerance_feasibility': 'FeasibilityTol',
                      'tolerance_markowitz': 'MarkowitzTol',
                      'tolerance_optimality': 'OptimalityTol',
                      'iteration_limit': 'IterationLimit',
                      'tolerance_barrier': 'BarConvTol',
                      'tolerance_integer': 'IntFeasTol',
                      'MIP_gap_abs': 'MIPGapAbs',
                      'MIP_gap': 'MIPGap'}
# http://www.gurobi.com/documentation/5./6/reference-manual/method
METHODS = {"auto": -1, "primal": 0, "dual": 1, "barrier": 2,
           "concurrent": 3, "deterministic concurrent": 4}
variable_kind_dict = {'continuous': GRB.CONTINUOUS, 'integer': GRB.INTEGER}
sense_dict = {'E': GRB.EQUAL, 'L': GRB.LESS_EQUAL, 'G': GRB.GREATER_EQUAL}
objective_senses = {'maximize': GRB.MAXIMIZE, 'minimize': GRB.MINIMIZE}
status_dict = {GRB.OPTIMAL: 'optimal', GRB.INFEASIBLE: 'infeasible',
               GRB.UNBOUNDED: 'unbounded', GRB.TIME_LIMIT: 'time_limit'}

def get_status(lp):
    status = lp.status
    if status in status_dict:
        status = status_dict[status]
    else:
        status = 'failed'
    return status

def get_objective_value(lp):
    return lp.ObjVal

def format_solution(lp, cobra_model, **kwargs):
    status = get_status(lp)
    if status not in ('optimal', 'time_limit'):
        the_solution = Solution(None, status=status)
    else:
        objective_value = lp.ObjVal
        x = [v.X for v in lp.getVars()]
        x_dict = {r.id: value for r, value in izip(cobra_model.reactions, x)}
        if lp.isMIP:
            y = y_dict = None  # MIP's don't have duals
        else:
            y = [c.Pi for c in lp.getConstrs()]
            y_dict = {m.id: value for m, value in izip(cobra_model.metabolites, y)}
        the_solution = Solution(objective_value, x=x, x_dict=x_dict, y=y,
                                y_dict=y_dict, status=status)
    return(the_solution)

def set_parameter(lp, parameter_name, parameter_value):
    if parameter_name == 'ModelSense' or parameter_name == "objective_sense":
        lp.setAttr('ModelSense', objective_senses[parameter_value])
    elif parameter_name == 'reuse_basis' and not parameter_value:
        lp.reset()
    else:
        parameter_name = parameter_mappings.get(parameter_name, parameter_name)
        if parameter_name == "Method" and isinstance(parameter_value,
                                                     string_types):
            parameter_value = METHODS[parameter_value]
        lp.setParam(parameter_name, parameter_value)

def change_variable_bounds(lp, index, lower_bound, upper_bound):
    variable = lp.getVarByName(str(index))
    variable.lb = lower_bound
    variable.ub = upper_bound


def change_variable_objective(lp, index, objective):
    variable = lp.getVarByName(str(index))
    variable.obj = objective


def change_coefficient(lp, met_index, rxn_index, value):
    met = lp.getConstrByName(str(met_index))
    rxn = lp.getVarByName(str(rxn_index))
    lp.chgCoeff(met, rxn, value)


def update_problem(lp, cobra_model, **kwargs):
    """A performance tunable method for updating a model problem file

    lp: A gurobi problem object

    cobra_model: the cobra.Model corresponding to 'lp'

    """
    #When reusing the basis only assume that the objective coefficients or bounds can change
    try:
        quadratic_component = kwargs['quadratic_component']
        if quadratic_component is not None:
            warn("update_problem does not yet take quadratic_component as a parameter")
    except:
        quadratic_component = None

    if 'copy_problem' in kwargs and kwargs['copy_problem']:
        lp = lp.copy()
    if 'reuse_basis' in kwargs and not kwargs['reuse_basis']:
        lp.reset()
    for the_variable, the_reaction in zip(lp.getVars(),
                                          cobra_model.reactions):
        the_variable.lb = float(the_reaction.lower_bound)
        the_variable.ub = float(the_reaction.upper_bound)
        the_variable.obj = float(the_reaction.objective_coefficient)


def create_problem(cobra_model, quadratic_component=None, **kwargs):
    """Solver-specific method for constructing a solver problem from
    a cobra.Model.  This can be tuned for performance using kwargs


    """
    lp = Model("")

    the_parameters = parameter_defaults
    if kwargs:
        the_parameters = parameter_defaults.copy()
        the_parameters.update(kwargs)

    for k, v in the_parameters.iteritems():
        set_parameter(lp, k, v)


    # Create variables
    #TODO:  Speed this up
    variable_list = [lp.addVar(float(x.lower_bound),
                               float(x.upper_bound),
                               float(x.objective_coefficient),
                               variable_kind_dict[x.variable_kind],
                               str(i))
                     for i, x in enumerate(cobra_model.reactions)]
    reaction_to_variable = dict(zip(cobra_model.reactions,
                                    variable_list))
    # Integrate new variables
    lp.update()

    #Constraints are based on mass balance
    #Construct the lin expression lists and then add
    #TODO: Speed this up as it takes about .18 seconds
    #HERE
    for i, the_metabolite in enumerate(cobra_model.metabolites):
        constraint_coefficients = []
        constraint_variables = []
        for the_reaction in the_metabolite._reaction:
            constraint_coefficients.append(the_reaction._metabolites[the_metabolite])
            constraint_variables.append(reaction_to_variable[the_reaction])
        #Add the metabolite to the problem
        lp.addConstr(LinExpr(constraint_coefficients, constraint_variables),
                     sense_dict[the_metabolite._constraint_sense.upper()],
                     the_metabolite._bound,
                     str(i))

    # Set objective to quadratic program
    if quadratic_component is not None:
        set_quadratic_objective(lp, quadratic_component)

    lp.update()
    return(lp)


def set_quadratic_objective(lp, quadratic_objective):
    if not hasattr(quadratic_objective, 'todok'):
        raise Exception('quadratic component must have method todok')
    variable_list = lp.getVars()
    linear_objective = lp.getObjective()
    # If there already was a quadratic expression set, this will be quadratic
    # and we need to extract the linear component
    if hasattr(linear_objective, "getLinExpr"):  # duck typing
        linear_objective = linear_objective.getLinExpr()
    gur_quadratic_objective = QuadExpr()
    for (index_0, index_1), the_value in quadratic_objective.todok().items():
        # gurobi does not multiply by 1/2 (only does v^T Q v)
        gur_quadratic_objective.addTerms(the_value * 0.5,
                                         variable_list[index_0],
                                         variable_list[index_1])
    # this adds to the existing quadratic objectives
    lp.setObjective(gur_quadratic_objective + linear_objective)

def solve_problem(lp, **kwargs):
    """A performance tunable method for updating a model problem file

    """
    #Update parameter settings if provided
    for k, v in kwargs.iteritems():
        set_parameter(lp, k, v)

    lp.update()
    lp.optimize()
    status = get_status(lp)
    return status

    
def solve(cobra_model, **kwargs):
    """

    """
    for i in ["new_objective", "update_problem", "the_problem"]:
        if i in kwargs:
            raise Exception("Option %s removed" % i)
    if 'error_reporting' in kwargs:
        warn("error_reporting deprecated")
        kwargs.pop('error_reporting')

    #Create a new problem
    lp = create_problem(cobra_model, **kwargs)

    ###Try to solve the problem using other methods if the first method doesn't work
    try:
        lp_method = kwargs['lp_method']
    except:
        lp_method = 0
    the_methods = [0, 2, 1]
    if lp_method in the_methods:
        the_methods.remove(lp_method)
    #Start with the user specified method
    the_methods.insert(0, lp_method)
    for the_method in the_methods:
        try:
            status = solve_problem(lp, lp_method=the_method)
        except:
            status = 'failed'
        if status == 'optimal':
            break

    return format_solution(lp, cobra_model)



#cobra.mlab.py
#System modules
import cPickle, numpy, os, cobra, warnings
from copy import deepcopy
from scipy.sparse import dok_matrix
from .. import Model, Reaction, Metabolite, Formula, ArrayBasedModel

from mlabwrap import mlab as matlab

# give an interface to the mlab system through cobra.matlab
cobra.matlab = matlab
matlab.__doc__ = """
This is an mlabwrap connection to MATLAB which can be used to call
MATLAB functions. For example, if model is a python model, the following
can be used to optimize the model in MATLAB:
> matlab_model = cobra.mlab.cobra_model_object_to_cobra_matlab_struct(model)
> result = cobra.matlab.optimizeCbModel(matlab_model)

Any MATLAB function can be called this way"""

# add path with module's python scripts to the MATLAB path
mlab_path = os.path.join(cobra.__path__[0], 'mlab', 'matlab_scripts')
matlab.addpath(mlab_path)

_possible_cobra_locations = ["~/MATLAB/cobra", "~/cobra",
    "~/Documents/MATLAB/cobra", "~/Documents/opencobra/matlab/cobra"]

def init_matlab_toolbox(matlab_cobra_path=None, discover_functions=True):
    """initialize the matlab cobra toolbox, and load its functions
    into mlab's namespace (very useful for ipython tab completion)

    matlab_cobra_path: the path to the directory containing the MATLAB
    cobra installation. Using the default None will attempt to find the
    toolbox in the MATLAB path
    
    discover_functions: Whether mlabwrap should autodiscover all cobra toolbox
    functions in matlab. This is convenient for tab completion, but may take
    some time."""
    # find initCobraToolbox.m
    if matlab_cobra_path is None:
        matlab_cobra_path = os.path.split(
            cobra.matlab.which('initCobraToolbox'))[0]
        if matlab_cobra_path == "":
            for i in _possible_cobra_locations:
                poss_path = os.path.expanduser(i)
                if os.path.isfile(poss_path + "/initCobraToolbox.m"):
                    matlab_cobra_path = poss_path
                    break
    if not os.path.isfile(os.path.join(matlab_cobra_path,
                           "initCobraToolbox.m")):
        print "initCobraToolbox not found in given path"
        return
    # store the current directory so we can return to it
    curdir = os.path.abspath(os.curdir)
    os.chdir(matlab_cobra_path)
    matlab.initCobraToolbox()
    os.chdir(curdir)
    # discover MATLAB functions
    for matlab_func in ["plot", "figure", "hist", "load", "find"]:
        getattr(matlab, matlab_func)
    for filename in os.listdir(mlab_path):
        getattr(matlab, filename.replace(".m", ""))
    if not discover_functions:
        return  # done now, do not need to find matlab cobra functions
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        walker = os.walk(matlab_cobra_path)
        # skip docs and external
        dirpath, dirnames, filenames = walker.next()
        dirnames.pop(dirnames.index("external"))
        dirnames.pop(dirnames.index("docs"))
        for root, dirs, files in walker:
            for filename in files:
                if filename.endswith(".m"):
                    try:
                        getattr(matlab, filename.replace(".m", ""))
                    except:
                        continue


#Project defined modules
def matlab_cell_to_python_list(the_cell):
    try:
        import mlabwrap
    except ImportError:
        print 'Could not import mlabwrap: ' + ImportError + '\n'
        return False
    #the_list = mlabwrap.mlab.cell_to_string(the_cell).rstrip('\t').split('\t')
    #Because the mlabwrap.mlab.cell_to_string function will end with a tab, it must
    #be removed.  Use indexing instead of rstrip because rstrip will mess up
    #when the last element in the string is empty.
    the_list = mlabwrap.mlab.cell_to_string(the_cell).split('\t')
    return the_list[:len(the_list) - 1]

def matlab_logical_to_python_logical(the_logical):
    try:
        import mlabwrap
    except ImportError:
        print 'Could not import mlabwrap: ' + ImportError + '\n'
        return False
    if mlabwrap.mlab.double(the_logical)[0][0] == 1:
        return True
    else:
        return False

def python_list_to_matlab_cell(the_list, transpose = False):
    try:
        import mlabwrap
    except ImportError:
        print 'Could not load mlabwrap ' + ImportError + '\n'
        return False
    if hasattr(the_list[0], 'id'):
        the_list = [x.id for x in the_list]
    #Remove the single apostrophes because matlab or mlabwrap has
    #problems with them.
    the_list = [x.replace(r"'",'') for x in the_list]
    the_string = '{' + str(the_list).lstrip('[').rstrip(']') + '}'
    if transpose:
        the_string = the_string.replace(',',';')
    return mlabwrap.mlab.eval(the_string)

def matlab_sparse_to_scipy_sparse(matlab_sparse_matrix):
    try:
        import mlabwrap
        import scipy.sparse
    except ImportError:
        print 'Could not load mlabwrap or scipy ' + ImportError + '\n'
        return False
    [row_indices, column_indices, the_values] = mlabwrap.mlab.find(matlab_sparse_matrix , nout = 3)
    #Change to 0-based indices
    row_indices = row_indices - 1
    column_indices = column_indices - 1
    the_shape = mlabwrap.mlab.size(matlab_sparse_matrix)
    the_shape =  tuple(map(int, the_shape[0,]))
    #This could be sped up by using a dok_matrix and zipping the indices and values
    #scipy_sparse_matrix = scipy.sparse_dok.matrix(the_shape)
    #scipy_sparse_matrix.update(zip(zip(row_indices, column_indices), the_values))
    #return scipy_sparse_matrix.tolil()
    scipy_sparse_matrix = scipy.sparse.lil_matrix(the_shape)
    for i in range(len(the_values)):
        scipy_sparse_matrix[int(row_indices[i][0]), int(column_indices[i][0])]  = the_values[i][0]
    return scipy_sparse_matrix

def scipy_sparse_to_mlab_sparse(scipy_sparse_matrix):
    """A more efficient method is needed for when the
    matrix is so big that making a dense version
    is a waste of computer effort."""
    try:
        import mlabwrap
    except ImportError:
        print 'Could not load mlabwrap ' + ImportError + '\n'
        return False
    return mlabwrap.mlab.sparse(scipy_sparse_matrix.todense())

def matlab_sparse_to_numpy_array(matlab_sparse_matrix):
    try:
        import mlabwrap
        import numpy
    except ImportError:
        print 'Could not load mlabwrap or numpy ' + ImportError + '\n'
        return False
    [row_indices, column_indices, the_values] = mlabwrap.mlab.find(matlab_sparse_matrix , nout = 3)
    #Change to 0-based indices
    row_indices = row_indices - 1
    column_indices = column_indices - 1
    the_shape = mlabwrap.mlab.size(matlab_sparse_matrix)
    numpy_array = numpy.zeros(the_shape[0,])
    for i in range(len(the_values)):
        numpy_array[int(row_indices[i][0]), int(column_indices[i][0])] = the_values[i][0]
    return numpy_array

def numpy_array_to_mlab_sparse(numpy_array):
    """A more efficient method is needed for when the
    matrix is so big that making a dense version
    is a waste of computer effort."""
    try:
        import mlabwrap
    except ImportError:
        print 'Could not load mlabwrap ' + ImportError + '\n'
        return False

    return mlabwrap.mlab.sparse(numpy_array)

def matlab_cobra_struct_to_python_cobra_object(matlab_struct):
    """Converts a COBRA toolbox 2.0 struct into a cobra.Model object using the mlabwrap matlab proxy

    """

    from cobra import Model
    try:
        from mlabwrap import mlab as matlab
    except:
        raise Exception('mlabwrap and MATLAB are required to use these functions. '+\
                        'They only function on Mac OS X and GNU/Linux')
    from copy import deepcopy
    from numpy import array
    from cobra.mlab import matlab_cell_to_python_list, matlab_sparse_to_scipy_sparse
    struct_fields = matlab_cell_to_python_list(matlab.fields(matlab_struct, nout=1))
    _S_dok = dok_matrix(matlab_sparse_to_scipy_sparse(matlab_struct.S))
    model_id = 'Matlab_Import'
    if 'description' in struct_fields:
        model_id = str(deepcopy(matlab_struct.description))
    the_model = Model(model_id)
    #Metabolite section
    #Deal with lack of compartment support in the COBRA Toolbox
    the_metabolites = matlab_cell_to_python_list(matlab_struct.mets)
    cobra_metabolites = []
    if the_metabolites[0].endswith(']'):
        [cobra_metabolites.append(Metabolite('%s_%s'%(x[:-3], x[-2]),
                                              compartment=x[-2]))
         for x in the_metabolites]
    elif the_metabolites[0][-2] == '_':
        [cobra_metabolites.append(Metabolite('%s_%s'%(x[:-2],x[-1]),
                                             compartment=x[-1]))
         for x in the_metabolites]
    else:
        raise Exception("Don't know how to extract compartment from %s"%the_metabolites[0])
    _b = deepcopy(matlab_struct.b).tolist()
    if 'metNames' in struct_fields:
        _metabolite_names = matlab_cell_to_python_list(matlab_struct.metNames)
    else:
        _metabolite_names = ['']*len(cobra_metabolites)
        
    if 'csense' in struct_fields:
        if isinstance(matlab_struct.csense, str):
            the_csense = matlab_struct.csense
            _constraint_sense = [the_csense[i] for i in range(len(the_csense))]
        else:
            _constraint_sense = matlab_cell_to_python_list(matlab_struct.csense)
    else:
        _constraint_sense = ['E']*len(cobra_metabolites)
    if 'metFormulas' in struct_fields:
        _metabolite_formulas = matlab_cell_to_python_list(matlab_struct.metFormulas)
    else:
        _metabolite_formulas = ['']*len(the_model.metabolites)
    if 'metCharge' in struct_fields:
        _metabolite_charges = matlab_struct.metCharge.flatten().tolist()
    else:
        _metabolite_charges = [None]*len(cobra_metabolites)
    if 'metCASID' in struct_fields:
        _metabolite_cas_id = matlab_cell_to_python_list(matlab_struct.metCASID)
    else:
        _metabolite_cas_id = [None]*len(cobra_metabolites)
    if 'metKeggID' in struct_fields:
        _metabolite_kegg_id = matlab_cell_to_python_list(matlab_struct.metKeggID)
    else:
        _metabolite_kegg_id = [None]*len(cobra_metabolites)
    the_compartments = {}
    for the_metabolite, b, n, c, f, ch, cas, kegg in zip(cobra_metabolites,
                                                         _b,
                                                         _metabolite_names,
                                                         _constraint_sense,
                                                         _metabolite_formulas,
                                                         _metabolite_charges,
                                                         _metabolite_cas_id,
                                                         _metabolite_kegg_id):
        the_metabolite._bound = b[0]
        the_metabolite.name = n
        the_metabolite._constraint_sense = c
        the_metabolite.formula = Formula(f)

        the_compartments[the_metabolite.compartment] = the_metabolite.compartment
        if ch is not None:
            the_metabolite.charge = the_metabolite.notes['CHARGE'] = int(ch)
        if cas is not None and cas != '':
            the_metabolite.notes['CASID'] = cas
        if kegg is not None and kegg != '':
            the_metabolite.notes['KEGGID'] = kegg
            
    metabolite_dict = dict([(x.id, x)
                            for x in cobra_metabolites])
    #Reaction section
    cobra_reactions = map(Reaction, matlab_cell_to_python_list(matlab_struct.rxns))
    _objective_coefficients = deepcopy(matlab_struct.c).tolist()
    _lower_bounds = deepcopy(matlab_struct.lb).tolist()
    _upper_bounds = deepcopy(matlab_struct.ub).tolist()
    _reversibility = deepcopy(matlab_struct.rev).tolist()
    if 'rxnNames' in struct_fields:
        _reaction_names = matlab_cell_to_python_list(matlab_struct.rxnNames)
    else:
        _reaction_names = ['']*len(the_model.reactions)
    if 'grRules' in struct_fields:
        _gene_reaction_rules = matlab_cell_to_python_list(matlab_struct.grRules)
    else:
        _gene_reaction_rules = ['']*len(the_model.reactions)
    if 'subSystems' in struct_fields:
        _subsystems = matlab_cell_to_python_list(matlab_struct.subSystems)
    else:
        _subsystems = ['']*len(the_model.reactions)
    for the_reaction, o, l, u, n, r, s, rev in zip(cobra_reactions,
                                              _objective_coefficients,
                                              _lower_bounds,
                                              _upper_bounds,
                                              _reaction_names,
                                              _gene_reaction_rules,
                                              _subsystems, _reversibility ):
        the_reaction.objective_coefficient = o[0]
        the_reaction.lower_bound = l[0]
        the_reaction.upper_bound = u[0]
        the_reaction.name = n
        the_reaction.gene_reaction_rule = r
        the_reaction.subsystem = s
        the_reaction.reversibility = rev[0]
        the_reaction.parse_gene_association()
    index_to_reaction = dict(zip(range(len(cobra_reactions)),
                                 cobra_reactions))
    index_to_metabolite = dict(zip(range(len(cobra_metabolites)),
                                 cobra_metabolites))
    for k, coefficient in _S_dok.items():
        the_reaction = index_to_reaction[k[1]]
        the_metabolite = index_to_metabolite[k[0]]
        the_reaction.add_metabolites({the_metabolite: coefficient})
    #solution needs to be redesigned from a matlab proxy.
    #Now populate all of the reactions
    #NOTE: Having a metabolite referenece here would be a good idea
    compartment_dict = {'e': 'extracellular',
                        'c': 'cytosol',
                        'p': 'periplasm'}
    for the_key in the_compartments:
        if the_key in compartment_dict:
            the_compartments[the_key] = compartment_dict[the_key]
    the_model.add_reactions(cobra_reactions)
    the_model.compartments = the_compartments
    the_model.solution = None
    return the_model


def cobra_model_object_to_cobra_matlab_struct(cobra_model):
    """This function converts all of the  object values to the
    corresponding model fields to update the mlab model.

    """
    try:
        from mlabwrap import mlab as matlab
    except:
        raise Exception('mlabwrap and MATLAB are required to use these functions. '+\
                        'They only function on Mac OS X and GNU/Linux')
    from cobra.mlab import python_list_to_matlab_cell, scipy_sparse_to_mlab_sparse
    if hasattr(cobra_model, 'to_array_based_model'):
        cobra_model = cobra_model.to_array_based_model()
    else:
        cobra_model.update()
    matlab_struct = matlab.struct()
    #Things that need a conversion:  S, rxnGeneMat,
    matlab_struct.mets = python_list_to_matlab_cell([x.id for x in cobra_model.metabolites],
                                                    transpose = True)
    matlab_struct.metNames =  python_list_to_matlab_cell([x.name for x in cobra_model.metabolites],
                                                         transpose = True)
    matlab_struct.metFormulas = python_list_to_matlab_cell([x.formula
                                                            for x in cobra_model.metabolites],
                                                           transpose = True)
    matlab_struct.genes = python_list_to_matlab_cell(cobra_model.genes, transpose = True)
    matlab_struct.grRules = python_list_to_matlab_cell([x.gene_reaction_rule
                                                        for x in cobra_model.reactions],
                                                       transpose = True)
    matlab_struct.rxns = python_list_to_matlab_cell([x.id for x in cobra_model.reactions],
                                                    transpose = True)
    matlab_struct.rxnNames = python_list_to_matlab_cell([x.name
                                                         for x in cobra_model.reactions],
                                                        transpose = True)
    matlab_struct.subSystems = python_list_to_matlab_cell([x.subsystem
                                                           for x in cobra_model.reactions],
                                                           transpose=True)

    if hasattr(cobra_model, 'constraint_sense'):
        matlab_struct.csense = reduce(lambda x,y: x+y, cobra_model.constraint_sense)
    #matlab_struct.csense = python_list_to_matlab_cell(['E']*len(cobra_model.metabolites), transpose = True)
    #TODO: inefficient conversion but who cares? matlab's on its way out
    matlab_struct.S = scipy_sparse_to_mlab_sparse(cobra_model.S)
    #Things that can be directly copied
    matlab_struct.b = cobra_model.b
    matlab_struct.c = cobra_model.objective_coefficients
    matlab_struct.lb = cobra_model.lower_bounds
    matlab_struct.ub = cobra_model.upper_bounds
    matlab_struct.rev = [x.reversibility for x in cobra_model.reactions]
    matlab_struct.description = cobra_model.description
    return(matlab_struct)



if __name__ == '__main__':
    from cPickle import load
    from time import time
    from numpy import round
    from cobra.manipulation import initialize_growth_medium
    try:
        matlab.changeCobraSolver('glpk','LP')
    except AttributeError:
        raise Exception('Could not run matlab function changeCobraSolver.  Is the ' +\
                        'COBRA Toolbox in your MATLAB path?')
    test_directory = '../test/data/'
    with open(test_directory + 'salmonella.pickle') as in_file:
        cobra_model = load(in_file)
    initialize_growth_medium(cobra_model, 'LPM')
    py_cobra_solution = repr(cobra_model.solution.f)
    matlab_struct = cobra_model_object_to_cobra_matlab_struct(cobra_model)

    matlab_result = matlab.optimizeCbModel(matlab_struct)
    matlab_solution = repr(float(matlab_result.f))
    if py_cobra_solution[:4] == matlab_solution[:4]:
        print 'SUCCESS: growth rate match between pyCOBRA and COBRA Toolbox: %s ~ %s'%(py_cobra_solution,
                                                                                       matlab_solution)
    else:
        print 'FAILURE: pyCOBRA and COBRA Toolbox do not match: %s !~ %s'%(py_cobra_solution,
                                                                                       matlab_solution)



#cobra.topology.reporter_metabolites.py: Module for topological analysis of cobra_models
#Based on Patil et al 2005 PNAS 102:2685-9
#TODO: Validate cobra.core compliance
from copy import deepcopy
from numpy import array, corrcoef, mean, std, tril, where, unique, zeros
from scipy.stats import norm, randint
from collections import defaultdict

def identify_reporter_metabolites(cobra_model, reaction_scores_dict,
                                  number_of_randomizations=1000, number_of_layers=1,
                                  scoring_metric='default', score_type='p',
                                  entire_network=False, background_correction=True,
                                  ignore_external_boundary_reactions=False):
    """Calculate the aggregate Z-score for the metabolites in the model.
    Ignore reactions that are solely spontaneous or orphan. Allow the scores to
    have multiple columns / experiments.   This will change the way the output
    is represented.

    cobra_model: A cobra.Model object

    TODO: CHANGE TO USING DICTIONARIES for the_reactions: the_scores

    reaction_scores_dict:  A dictionary where the keys are reactions in cobra_model.reactions
    and the values are the scores.  Currently, only supports a single numeric value as
    the value; however, this will be updated to allow for lists

    number_of_randomizations: Integer.  Number of random shuffles of the
    scores to assess which are significant.

    number_of_layers: 1 is the only option supported
    
    scoring_metric:  default means divide by k**0.5

    score_type: 'p' Is the only option at the moment and indicates p-value.

    entire_network: Boolean.  Currently, only compares scores calculated from the_reactions

    background_correction: Boolean.  If True apply background correction to the aggreagate
    Z-score

    ignore_external_boundary_reactions: Not yet implemented. Boolean.  If True do not count exchange reactions when
    calculating the score.

    

    
    """
    #Add in a function to calculate based on correlation coefficients and to
    #deal with other multidimensional data. 
    the_reactions = reaction_scores_dict.keys()
    the_scores = reaction_scores_dict.values()
    if score_type == 'p' and not hasattr(the_scores[0], '__iter__'):
        #minimum and maximum p-values are used to prevent numerical problems.
        #haven't decided whether an arbitrary min / max 1e-15 is preferred to
        #blunting the ends based on the values closest to 0 or 1.
        the_reactions = reaction_scores_dict.keys()
        the_scores = array(reaction_scores_dict.values())
        minimum_p = min(the_scores[the_scores.nonzero()[0]])
        maximum_p = max(the_scores[where(the_scores < 1)[0]])
        the_scores[where(the_scores < minimum_p)] = minimum_p
        the_scores[where(the_scores > maximum_p)] = maximum_p
        the_scores = -norm.ppf(the_scores)
        #update the dictionary with the new scores
        reaction_scores_dict = dict(zip(the_reactions, the_scores))
    elif hasattr(the_scores[0], '__iter__'):
        #In the case that the_scores is a list of lists, assume that each list is
        #the score for each reaction in the_reactions across all reactions.  Then
        #for each metabolite, calculate the invnorm(|Pearson Correlation
        #Coefficient| for each reaction pair that it links.
        raise Exception("This isn't implemented yet")
    
    #Get the connectivity for each metabolite
    the_metabolites = set()
    [the_metabolites.update(x._metabolites)
     for x in reaction_scores_dict];

    metabolite_scores = {}
    metabolite_connections = {}
    #Calculate the score for each metabolite
    for the_metabolite in the_metabolites:
        nonspontaneous_connections = [x for x in the_metabolite._reaction
                                      if x.gene_reaction_rule.lower() not in
                                      ['s0001', '']]
        tmp_score = 0
        number_of_connections = len(nonspontaneous_connections)
        for the_reaction in nonspontaneous_connections:
            if the_reaction not in reaction_scores_dict:
                if not entire_network:
                    number_of_connections -= 1
                continue
            else:
                tmp_score += reaction_scores_dict[the_reaction]
        metabolite_scores[the_metabolite] = tmp_score
        metabolite_connections[the_metabolite] = number_of_connections

    #NOTE: Doing the corrections based only on the significantly perturbed scores
    #is probably going to underestimate the significance.
    if background_correction:
        correction_dict = {}
        for i in set(metabolite_connections.values()):
            #if entire_network # add in a section to deal with the situation where
            #the entire network structure is considered by only have p-values for
            #a limited subset.
            #
            #Basically, what we're doing here is that for each i we select i
            #scores number_of_randomizations times
            the_random_indices = randint.rvs(0,len(the_scores), size=(number_of_randomizations, i))
            random_score_distribution = array([sum(the_scores[x]) for x in list(the_random_indices)]) /i**0.5
            correction_dict[i] = [mean(random_score_distribution),
                                      std(random_score_distribution,ddof=1)] 

    for the_metabolite, the_score in metabolite_scores.iteritems():
        number_of_connections = metabolite_connections[the_metabolite]
        if number_of_connections > 0:
            #Correct based on background distribution
            if background_correction:
                #if the list of scores is only for significant perturbations then the
                #background correction shouldn't be applied because the current sampling
                #method only takes into account the_scores not the entire network.
                #It'd be more accurate to assign unscored reactions a default score.
                the_score = ((the_score / number_of_connections**.5) -
                             correction_dict[number_of_connections][0]) / \
                             correction_dict[number_of_connections][1]
            else:
                the_score = the_score / number_of_connections**.5
            #Update the score
            metabolite_scores[the_metabolite] = the_score



    return_dictionary = {'scores': metabolite_scores,
                         'connections': metabolite_connections}
    if background_correction:
        return_dictionary['corrections'] = correction_dict

    return(return_dictionary)
    
def ppmap_identify_reporter_metabolites(keywords):
    """
    A function that receives a dict with all of the parameters for identify_reporter_metabolites
    Serves to make it possible to call the reporter metabolites function from ppmap.
    It only will be useful for parallel experiments not for breaking up a single experiment.
    
    """
    the_results = identify_reporter_metabolites(**keywords)
    return({'id': the_id, 'results': the_results })

if __name__ == '__main__':
    from cPickle import load
    from time import time
    solver = 'glpk'
    from cobra.test import salmonella_pickle, salmonella_reaction_p_values_pickle
    with open(salmonella_pickle) as in_file:
        cobra_model = load(in_file)
    with open(salmonella_reaction_p_values_pickle) as in_file:
        reaction_p = load(in_file)

    the_reactions = map(cobra_model.reactions.get_by_id, reaction_p.keys())
    the_scores = reaction_p.values()
    reaction_scores_dict = dict(zip(the_reactions, the_scores))
    
    tmp_reps = identify_reporter_metabolites(cobra_model, reaction_scores_dict,
                                             background_correction=True)

    print 'Need to add in validation for the test'



from warnings import warn
from copy import deepcopy
from .Formula import Formula
from .Object import Object


class Species(Object):
    """Species is a class for holding information regarding
    a chemical Species


    """

    def __init__(self, id=None, formula=None,
                 name=None, compartment=None, mnx_id=None):
        """
        id: A string.

        formula: cobra.Formula or str of a chemical formula.  Defaults to None
        to save time in pickling and such.

        name: String.  A human readable name.

        compartment: None or a dictionary indicating the cellular location
        of the metabolite.  Used when in a cobra.Reaction or Model
        object

        mnx_id: None or a String of the MetaNetX.org ID for the object.

        """
        Object.__init__(self, id, mnx_id=mnx_id)
        self.name = name
        if not name:
            self.name = self.id
        if isinstance(formula, str):
            formula = Formula(formula)

        self.formula = formula
        self.parse_composition()
        # because in a Model a metabolite may participate in multiple Reactions
        self.compartment = compartment
        # self.model is None or refers to the cobra.Model that
        # contains self
        self._model = self.charge = None
        # references to reactions that operate on this species
        self._reaction = set()

    @property
    def reactions(self):
        return frozenset(self._reaction)

    def parse_composition(self):
        """Breaks the chemical formula down by element.
        Useful for making sure Reactions are balanced.'

        """
        if isinstance(self.formula, Formula):
            self.formula.parse_composition()
        elif isinstance(self.formula, str):
            self.formula = Formula(self.formula)

    def __getstate__(self):
        """Remove the references to container reactions when serializing to
        avoid problems associated with recursion.

        """
        state = Object.__getstate__(self)
        state['_reaction'] = set()
        return state

    def copy(self):
        """When copying a reaction, it is necessary to deepcopy the
        components so the list references aren't carried over.

        Additionally, a copy of a reaction is no longer in a cobra.Model.

        This should be fixed with self.__deecopy__ if possible
        """
        return deepcopy(self)

    def guided_copy(self, the_model):
        """.. deprecated :: 0.3 Use copy directly"""
        warn("deprecated")
        the_copy = Object.guided_copy(self)
        # Copy the more complex objects in a faster fashion
        the_copy.formula = deepcopy(self.formula)
        the_copy._model = the_model
        the_copy._reaction = set()
        return(the_copy)

    def get_reaction(self):
        """Returns a list of Reactions that contain this Species"""
        warn("deprecated, used species.reactions instead")
        return list(self._reaction)

    def get_model(self):
        """Returns the Model object that contain this Object"""
        print("get_model is deprecated. used model property instead")
        return self._model

    @property
    def model(self):
        return(self._model)



#!/usr/bin/env python
# This script regenerates pickles of cobra Models.  Should be
# performed after updating core classes to prevent subtle bugs.
try:
    from cPickle import load, dump
except:
    from pickle import load, dump

from os.path import isfile

from cobra import Model
from cobra.version import get_version
from cobra.io import read_sbml_model, read_legacy_sbml, write_sbml_model
from cobra.io import save_matlab_model, save_json_model
from cobra.test import create_test_model

model_names = ['salmonella', 'iJO1366',]

for model_name in model_names:
    # read in old pickle and model from sbml
    model_pickle = model_name + '.pickle'
    if model_name == "iJO1366":
        new_model = read_legacy_sbml(model_name + '.xml')
    else:
        new_model = read_sbml_model(model_name + '.xml')
    # update other attributes
    if isfile(model_name + ".genes"):
        with open(model_name + ".genes", "rb") as infile:
            gene_names = load(infile)
        for gene in new_model.genes:
            gene.name = gene_names[gene.id]
    if isfile(model_name + ".media"):
        with open(model_name + ".media", "rb") as infile:
            new_model.media_compositions = load(infile)
    new_model._cobra_version = get_version()
    # write out new pickle
    with open(model_pickle, 'wb') as outfile:
        dump(new_model, outfile, protocol=2)
    # write out other formats for iJO1366
    if model_name == "iJO1366":
        save_matlab_model(new_model, model_name + ".mat")
        save_json_model(new_model, model_name + ".json")
    if model_name == "salmonella":
        write_sbml_model(new_model, model_name + "_fbc.xml")



try:
    import numpy
except:
    numpy = None

from .essentiality import assess_medium_component_essentiality
from .variability import flux_variability_analysis
from .single_deletion import single_deletion

if numpy:
    from .double_deletion import double_deletion
else:
    from warnings import warn
    warn("double_deletion requires numpy")
    del warn
del numpy



#!/usr/bin/env python

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
# 
#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials
#       provided with the distribution.
# 
#     * Neither the name of Kirk Strauser nor the names of other
#       contributors may be used to endorse or promote products
#       derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Very basic parallel processing support

Implements a work-alike of the builtin map() function that distributes
work across many processes.  As it uses Parallel Python to do the
actual multi-processing, code using this must conform to the usual PP
restrictions (arguments must be serializable, etc.)
"""

__author__  = "Kirk Strauser <kirk@strauser.com>"
__version__ = "$Rev: 1139 $"
__date__    = "$Date: 2008-04-16 $"

import time
import __builtin__

try:
    import pp
except:
    raise Exception('Could not import pp')
__STATE = {'server': None}

def ppmap(processes, function, sequence, *sequences):
    """Split the work of 'function' across the given number of
    processes.  Set 'processes' to None to let Parallel Python
    autodetect the number of children to use.

    Although the calling semantics should be identical to
    __builtin__.map (even using __builtin__.map to process
    arguments), it differs in that it returns a generator instead of a
    list.  This enables lazy evaluation of the results so that other
    work can be done while the subprocesses are still running.

    >>> def rangetotal(n): return n, sum(range(n))
    >>> list(map(rangetotal, range(1, 6)))
    [(1, 0), (2, 1), (3, 3), (4, 6), (5, 10)]
    >>> list(ppmap(1, rangetotal, range(1, 6)))
    [(1, 0), (2, 1), (3, 3), (4, 6), (5, 10)]
    """

    # Create a new server if one isn't already initialized
    if not __STATE['server']:
        __STATE['server'] = pp.Server()
    
    def submit(*args):
        """Send a job to the server"""
        return __STATE['server'].submit(function, args, globals=globals())

    # Merge all the passed-in argument lists together.  This is done
    # that way because as with the map() function, at least one list
    # is required but the rest are optional.
    a = [sequence]
    a.extend(sequences)
    available_processes = sum( __STATE['server'].get_active_nodes().values())
    # Set the requested level of multi-processing
    if available_processes < processes:
        __STATE['server'].set_ncpus(processes-available_processes or 'autodetect')

    # First, submit all the jobs.  Then harvest the results as they
    # come available.
    return (subproc() for subproc in __builtin__.map(submit, *a))

if __name__ == '__main__':
    def add(x, y, z):
        """Add three values"""
        return x + y + z

    def busybeaver(x):
        """This can take a while"""
        for num in range(1000000):
            x = x + num
        return x
    # Immediate evaluation example
    start = time.time()
    results = ppmap(None, busybeaver, range(10))
    print 'Time to queue the jobs:', time.time() - start
    start = time.time()
    # Casting the ppmap generator to a list forces each result to be
    # evaluated.  When done immediately after the jobs are submitted,
    # our program twiddles its thumbs while the work is finished.
    print list(results)
    print 'Time to get the results:', time.time() - start

    # Delayed evaluation example
    start = time.time()
    results = ppmap(None, busybeaver, range(10))
    print 'Time to queue the jobs:', time.time() - start
    # In contrast with the above example, this time we're submitting a
    # batch of jobs then going off to do more work while they're
    # processing.  Maybe "time.sleep" isn't the most exciting example,
    # but it illustrates the point that our main program can do work
    # before ppmap() is finished.  Imagine that you're submitting some
    # heavyweight image processing jobs at the beginning of your
    # program, going on to do other stuff like fetching more work to
    # do from a remote server, then coming back later to handle the
    # results.
    time.sleep(5)
    start = time.time()
    print list(results)
    print 'Time to get the first results:', time.time() - start

    # Built-in map example
    print map(add, [1, 2, 3], [4, 5, 6], [7, 8, 9])

    # Trivial ppmap tests
    for i in range(10):
        print '-' * 30
        start = time.time()
        print i, 'adders'
        print ppmap(i, add, [1, 2, 3], [4, 5, 6], [7, 8, 9])
        print 'Iteration time:', time.time() - start

    # Heavier ppmap tests
    for i in range(10):
        print '-' * 30
        start = time.time()
        print i, 'beavers'
        print ppmap(i, busybeaver, range(10))
        print 'Iteration time:', time.time() - start



##cobra.solvers.gurobi_solver
#Interface to the gurobi 5.0.1 python and java solvers
#QPs are not yet supported on java
from warnings import warn
from os import name as __name
from copy import deepcopy
###solver specific parameters
from .parameters import status_dict, variable_kind_dict, \
     sense_dict, parameter_mappings, parameter_defaults, \
     objective_senses, default_objective_sense

from ..core.Solution import Solution
from time import time
solver_name = 'gurobi'
objective_senses = objective_senses[solver_name]
parameter_mappings = parameter_mappings[solver_name]
parameter_defaults = parameter_defaults[solver_name]
#Functions that are different for java implementation of a solver
## from jarray import array as j_array
## def array(x, variable_type='d'):
##     return j_array(x, variable_type)

from gurobi import GRB
variable_kind_dict = eval(variable_kind_dict[solver_name])
status_dict = eval(status_dict[solver_name])

from gurobi import GRBModel, GRBEnv
from gurobi import GRBLinExpr
from gurobi import GRBQuadExpr as QuadExpr
__solver_class = GRBModel
#TODO: Create a pythonesqe class similar to in glpk_solver
def Model(name=''):
    grb_environment = GRBEnv(name)
    tmp_model = GRBModel(grb_environment)
    return tmp_model
def LinExpr(coefficients, variables):
    coefficients, variables = map(list, [coefficients, variables])
    tmp_expression = GRBLinExpr()
    tmp_expression.addTerms(coefficients, variables)
    return tmp_expression

def get_status(lp):
    status = lp.get(GRB.IntAttr.Status)
    if status in status_dict:
        status = status_dict[status]
    else:
        status = 'failed'
    return status

def set_parameter(lp, parameter_name, parameter_value):
    """Sets model parameters and attributes.
    
    """
    grb_environment = lp.getEnv()
    try:
        if hasattr(GRB.DoubleParam, parameter_name):
            grb_environment.set(eval('GRB.DoubleParam.%s'%parameter_name),
                                     parameter_value)
        elif hasattr(GRB.IntParam, parameter_name):
            grb_environment.set(eval('GRB.IntParam.%s'%parameter_name),
                                     parameter_value)
        elif hasattr(GRB.StringParam, parameter_name):
            grb_environment.set(eval('GRB.StringParam.%s'%parameter_name),
                                parameter_value)
        elif hasattr(GRB.IntAttr, parameter_name):
            if parameter_name == 'ModelSense':
                parameter_value = objective_senses[parameter_value]
            lp.set(eval('GRB.IntAttr.%s'%parameter_name),
                                parameter_value)
        else:
            warn("%s is not a DoubleParam, IntParam, StringParam, IntAttr"%parameter_name)
            ## raise Exception("%s is not a DoubleParam, IntParam, StringParam, IntAttr"%parameter_name)
    except Exception, e:
        warn("%s %s didn't work %s"%(parameter_name, parameter_value, e))

def get_objective_value(lp):
    return lp.get(GRB.DoubleAttr.ObjVal)

def format_solution(lp, cobra_model, **kwargs):
    """
    """
    status = get_status(lp)
    if status not in ('optimal', 'time_limit'):
        the_solution = Solution(None, status=status)
    else:
        x_dict = dict(((v.get(GRB.StringAttr.VarName),
                        v.get(GRB.DoubleAttr.X))
                       for v in lp.getVars()))
        x = [x_dict[v.id] for v in cobra_model.reactions]
        objective_value = lp.get(GRB.DoubleAttr.ObjVal)
        if lp.get(GRB.IntAttr.IsMIP) != 0:
            y = y_dict = None #MIP's don't have duals
        else:
            y_dict = dict(((c.get(GRB.StringAttr.ConstrName), c.get(GRB.DoubleAttr.Pi))
                          for c in lp.getConstrs()))
            y = list([y_dict[v.id] for v in cobra_model.metabolites])
        the_solution = Solution(objective_value, x=x, x_dict=x_dict, y=y,
                                y_dict=y_dict, status=status)
    return(the_solution)

def update_problem(lp, cobra_model, **kwargs):
    """A performance tunable method for updating a model problem file

    lp: A gurobi problem object

    cobra_model: the cobra.Model corresponding to 'lp'

    """
    #When reusing the basis only assume that the objective coefficients or bounds can change
    try:
        quadratic_component = kwargs['quadratic_component']
        if quadratic_component is not None:
            warn("update_problem does not yet take quadratic_component as a parameter")
    except:
        quadratic_component = None

    if 'reuse_basis' in kwargs and not kwargs['reuse_basis']:
        lp.reset()
    for the_variable, the_reaction in zip(lp.getVars(),
                                          cobra_model.reactions):
        the_variable.set(GRB.DoubleAttr.LB, float(the_reaction.lower_bound))
        the_variable.set(GRB.DoubleAttr.UB, float(the_reaction.upper_bound))
        the_variable.set(GRB.DoubleAttr.Obj, float(the_reaction.objective_coefficient))




###
sense_dict = eval(sense_dict[solver_name])
def create_problem(cobra_model,  **kwargs):
    """Solver-specific method for constructing a solver problem from
    a cobra.Model.  This can be tuned for performance using kwargs


    """
    lp = Model("")
    #Silence the solver
    set_parameter(lp, 'OutputFlag', 0)

    the_parameters = parameter_defaults
    if kwargs:
        the_parameters = deepcopy(parameter_defaults)
        the_parameters.update(kwargs)

    [set_parameter(lp, parameter_mappings[k], v)
         for k, v in the_parameters.iteritems() if k in parameter_mappings]
    quadratic_component = the_parameters['quadratic_component']
    objective_sense = objective_senses[the_parameters['objective_sense']]


    # Create variables
    #TODO:  Speed this up
    variable_list = [lp.addVar(float(x.lower_bound),
                               float(x.upper_bound),
                               float(x.objective_coefficient),
                               variable_kind_dict[x.variable_kind],
                               x.id)
                     for x in cobra_model.reactions]
    reaction_to_variable = dict(zip(cobra_model.reactions,
                                    variable_list))
    # Integrate new variables
    lp.update()
    #Set objective to quadratic program
    if quadratic_component is not None:
        if not hasattr(quadratic_component, 'todok'):
            raise Exception('quadratic component must have method todok')

        quadratic_objective = QuadExpr()
        for (index_0, index_1), the_value in quadratic_component.todok().items():
            quadratic_objective.addTerms(the_value,
                                   variable_list[index_0],
                                   variable_list[index_1])
        #Does this override the linear objective coefficients or integrate with them?
        lp.setObjective(quadratic_objective, sense=objective_sense)
    #Constraints are based on mass balance
    #Construct the lin expression lists and then add
    #TODO: Speed this up as it takes about .18 seconds
    #HERE
    for the_metabolite in cobra_model.metabolites:
        constraint_coefficients = []
        constraint_variables = []
        for the_reaction in the_metabolite._reaction:
            constraint_coefficients.append(the_reaction._metabolites[the_metabolite])
            constraint_variables.append(reaction_to_variable[the_reaction])
        #Add the metabolite to the problem
        lp.addConstr(LinExpr(constraint_coefficients, constraint_variables),
                     sense_dict[the_metabolite._constraint_sense.upper()],
                     the_metabolite._bound,
                     the_metabolite.id)



    return(lp)
###

###
def solve_problem(lp, **kwargs):
    """A performance tunable method for updating a model problem file

    """
    #Update parameter settings if provided
    if kwargs:
        [set_parameter(lp, parameter_mappings[k], v)
         for k, v in kwargs.iteritems() if k in parameter_mappings]

    try:
        print_solver_time = kwargs['print_solver_time']
        start_time = time()
    except:
        print_solver_time = False
    lp.update()
    #Different methods to try if lp_method fails
    lp.optimize()
    status = get_status(lp)
    if print_solver_time:
        print 'optimize time: %f'%(time() - start_time)
    return status

    
def solve(cobra_model, **kwargs):
    """

    """
    #Start out with default parameters and then modify if
    #new onese are provided
    the_parameters = deepcopy(parameter_defaults)
    if kwargs:
        the_parameters.update(kwargs)
    #Update objectives if they are new.
    if 'new_objective' in the_parameters and \
           the_parameters['new_objective'] not in ['update problem', None]:
       from ..flux_analysis.objective import update_objective
       update_objective(cobra_model, the_parameters['new_objective'])

    if 'the_problem' in the_parameters:
        the_problem = the_parameters['the_problem']
    else:
        the_problem = None
    if 'error_reporting' in the_parameters:
        error_reporting = the_parameters['error_reporting']
    else:
        error_reporting = False

    if isinstance(the_problem, __solver_class):
        #Update the problem with the current cobra_model
        lp = the_problem
        update_problem(lp, cobra_model, **the_parameters)
    else:
        #Create a new problem
        lp = create_problem(cobra_model, **the_parameters)
    #Deprecated way for returning a solver problem created from a cobra_model
    #without performing optimization
    if the_problem == 'setup':
            return lp

    ###Try to solve the problem using other methods if the first method doesn't work
    try:
        lp_method = the_parameters['lp_method']
    except:
        lp_method = 0
    the_methods = [0, 2, 1]
    if lp_method in the_methods:
        the_methods.remove(lp_method)
    #Start with the user specified method
    the_methods.insert(0, lp_method)
    for the_method in the_methods:
        the_parameters['lp_method'] = the_method
        try:
            status = solve_problem(lp, **the_parameters)
        except:
            status = 'failed'
        if status == 'optimal':
            break
    status = solve_problem(lp, **the_parameters)
    the_solution = format_solution(lp, cobra_model)
    if status != 'optimal' and error_reporting:
        print '%s failed: %s'%(solver_name, status)
    cobra_model.solution = the_solution
    solution = {'the_problem': lp, 'the_solution': the_solution}
    return solution



from __future__ import with_statement
#cobra.flux_analysis.essentiality.py
#runs flux variablity analysis on a Model object.
from os import name as __name
from warnings import warn
if __name == 'java':
    warn("moma is not supported on %s"%__name)
    warn("flux_analysis.double_deletion is not supported on %s"%__name)
    def moma(a, **kwargs):
        raise Exception("moma is not supported on %s"%__name)

    def double_deletion(a, **kwargs):
        raise Exception("flux_analysis.double_deletion is not supported on %s"%__name)
else:
    try:
        from cobra.flux_analysis.moma import moma
    except:
        warn("moma does not appear to be functional on your system")
from cobra.manipulation import initialize_growth_medium
def assess_medium_component_essentiality(cobra_model, the_components=None,
                                         the_medium=None, medium_compartment='e', solver='glpk',
                                         the_problem='return',
                                         the_condition=None, method='fba'):
    """Determines which components in an in silico medium are essential for growth in the
    context of the remaining components.

    cobra_model: A Model object.

    the_components: None or a list of external boundary reactions that will be sequentially
    disabled.

    the_medium: Is None, a string, or a dictionary.  If a string then the
    initialize_growth_medium function expects that the_model has an
    attribute dictionary called media_compositions, which is a dictionary of
    dictionaries for various medium compositions.  Where a medium
    composition is a dictionary of external boundary reaction ids for the medium
    components and the external boundary fluxes for each medium component.

    medium_compartment: the compartment in which the boundary reactions supplying the medium
    components exist

    NOTE: that these fluxes must be negative because the convention is backwards means something
    is feed into the system.

    solver: 'glpk', 'gurobi', or 'cplex'

    the_problem: Is None, 'return', or an LP model object for the solver.

    returns:
     essentiality_dict:  A dictionary providing the maximum growth rate accessible when
     the respective component is removed from the medium.

    """
    if method.lower() == 'moma':
        wt_model = cobra_model.copy()
    cobra_model = cobra_model.copy()

    if isinstance(the_medium, str):
        try:
            the_medium = cobra_model.media_compositions[the_medium]
        except:
            raise Exception(the_medium + " is not in cobra_model.media_compositions")
    if the_medium is not None:
        initialize_growth_medium(cobra_model, the_medium, medium_compartment)
        if the_components is None:
            the_components = the_medium.keys()
    if not the_components:
            raise Exception("You need to specify the_components or the_medium")
    essentiality_dict = {}
    for the_component in the_components:
        the_reaction = cobra_model.reactions.get_by_id(the_component)
        original_lower_bound = float(the_reaction.lower_bound)
        the_reaction.lower_bound = 0.
        if method.lower() == 'fba':
            cobra_model.optimize(solver=solver, the_problem=the_problem)
            objective_value = cobra_model.solution.f
        elif method.lower() == 'moma':
           objective_value = moma(wt_model, cobra_model, solver=solver)['objective_value'] 
        essentiality_dict[the_component] = objective_value
        the_reaction.lower_bound = original_lower_bound

    return(essentiality_dict)

def deletion_analysis(cobra_model, the_medium=None, deletion_type='single',
                      work_directory=None, growth_cutoff=0.001,
                      the_problem='return', number_of_processes=6, element_type='gene',
                      solver='glpk', error_reporting=None, method='fba', element_list=None):
    """Performs single and/or double deletion analysis on all the genes in the model.  Provides
    an interface to parallelize the deletion studies.

    """
    raise Exception("Deletion analysis has been dropped in favor of the single_deletion and double_deletion modules")



from copy import deepcopy
from warnings import warn

from .. import Reaction


def initialize_growth_medium(cobra_model, the_medium='MgM',
                             external_boundary_compartment='e',
                             external_boundary_reactions=None,
                             reaction_lower_bound=0.,
                             reaction_upper_bound=1000.,
                             irreversible=False,
                             reactions_to_disable=None):
    """Sets all of the input fluxes to the model to zero and then will
    initialize the input fluxes to the values specified in the_medium if
    it is a dict or will see if the model has a composition dict and use
    that to do the initialization.

    cobra_model: A cobra.Model object.


    the_medium: A string, or a dictionary.
    If a string then the initialize_growth_medium function expects that
    the_model has an attribute dictionary called media_compositions, which is a
    dictionary of dictionaries for various medium compositions.  Where a medium
    composition is a dictionary of external boundary reaction ids for the
    medium components and the external boundary fluxes for each medium
    component.


    external_boundary_compartment: None or a string.
    If not None then it specifies the compartment in which to disable all of
    the external systems boundaries.

    external_boundary_reactions: None or a list of external_boundaries that are
    to have their bounds reset.  This acts in conjunction with
    external_boundary_compartment.


    reaction_lower_bound: Float.  The default value to use for the lower
    bound for the boundary reactions.

    reaction_upper_bound: Float.  The default value to use for the upper
    bound for the boundary.

    irreversible: Boolean.  If the model is irreversible then the medium
    composition is taken as the upper bound

    reactions_to_disable: List of reactions for which the upper and lower
    bounds are disabled.  This is superceded by the contents of
    media_composition

    """
    # Zero all of the inputs to the model
    if hasattr(the_medium, 'keys'):
        medium_composition = the_medium
    else:
        if hasattr(cobra_model, 'media_compositions'):
            if the_medium in cobra_model.media_compositions:
                medium_composition = cobra_model.media_compositions[the_medium]
            else:
                raise Exception("%s is not in the model's media list" %
                                the_medium)
        else:
            raise Exception("the model doesn't have attribute "
                            "media_compositions and the medium is not a dict")
    if external_boundary_reactions is not None:
        if isinstance(external_boundary_reactions[0], str):
            external_boundary_reactions = map(cobra_model.reactions.get_by_id,
                                              external_boundary_reactions)
    elif external_boundary_compartment is None:
            warn("We are initializing the medium without first adjusting all"
                 "external boundary reactions")

    # Select the system_boundary reactions to reset
    if external_boundary_compartment is not None:
        _system_boundaries = dict([(x, x.get_compartments())
                                   for x in cobra_model.reactions
                                   if x.boundary == 'system_boundary'])
        [_system_boundaries.pop(k) for k, v in list(_system_boundaries.items())
         if len(v) == 1 and external_boundary_compartment not in v]
        if external_boundary_reactions is None:
            external_boundary_reactions = _system_boundaries.keys()
        else:
            external_boundary_reactions += _system_boundaries.keys()

    for the_reaction in external_boundary_reactions:
        the_reaction.lower_bound = reaction_lower_bound
        if the_reaction.upper_bound == 0:
            the_reaction.upper_bound = reaction_upper_bound
    # Disable specified reactions
    if reactions_to_disable is not None:
        if isinstance(reactions_to_disable[0], str):
            reactions_to_disable = map(cobra_model.reactions.get_by_id,
                                       reactions_to_disable)
        for the_reaction in reactions_to_disable:
            the_reaction.lower_bound = the_reaction.upper_bound = 0.

    # Update the model inputs based on the_medium
    for the_component in medium_composition.keys():
        the_reaction = cobra_model.reactions.get_by_id(the_component)
        if irreversible:
            the_reaction.upper_bound = medium_composition[the_component]
        else:
            the_reaction.lower_bound = medium_composition[the_component]


def convert_to_irreversible(cobra_model):
    """Split reversible reactions into two irreversible reactions

    These two reactions will proceed in opposite directions. This
    guarentees that all reactions in the model will only allow
    positive flux values, which is useful for some modeling problems.

    cobra_model: A Model object which will be modified in place.

    """
    reactions_to_add = []
    for reaction in cobra_model.reactions:
        # If a reaction is reverse only, the forward reaction (which
        # will be constrained to 0) will be left in the model.
        if reaction.lower_bound < 0:
            reverse_reaction = Reaction(reaction.id + "_reverse")
            reverse_reaction.lower_bound = 0
            reverse_reaction.upper_bound = reaction.lower_bound * -1
            reverse_reaction.objective_coefficient = \
                reaction.objective_coefficient * -1
            reaction.lower_bound = 0
            # Make the directions aware of each other
            reaction.reflection = reverse_reaction
            reverse_reaction.reflection = reaction
            reaction_dict = dict([(k, v*-1)
                                  for k, v in reaction._metabolites.items()])
            reverse_reaction.add_metabolites(reaction_dict)
            reverse_reaction._model = reaction._model
            reverse_reaction._genes = reaction._genes
            for gene in reaction._genes:
                gene._reaction.add(reverse_reaction)
            reverse_reaction._gene_reaction_rule = reaction._gene_reaction_rule
            reactions_to_add.append(reverse_reaction)
    cobra_model.add_reactions(reactions_to_add)


def revert_to_reversible(cobra_model, update_solution=True):
    """This function will convert a reversible model made by
    convert_to_irreversible into a reversible model.

    cobra_model: A cobra.Model which will be modified in place.

    """
    reverse_reactions = [x for x in cobra_model.reactions
                         if x.reflection is not None and
                         x.id.endswith('_reverse')]

    # If there are no reverse reactions, then there is nothing to do
    if len(reverse_reactions) == 0:
        return

    for reverse in reverse_reactions:
        forward = reverse.reflection
        forward.lower_bound = -reverse.upper_bound
        forward.reflection = None
    # Since the metabolites and genes are all still in
    # use we can do this faster removal step.  We can
    # probably speed things up here.
    cobra_model.remove_reactions(reverse_reactions)
    # fix the solution
    if update_solution and cobra_model.solution is not None and \
            cobra_model.solution.status != "NA":
        x_dict = cobra_model.solution.x_dict
        # Check if the model was optimized while it was still reversible. If so
        # then the solution does not need to be fixed.
        if reverse_reactions[0].id not in x_dict:
            return
        # Update x and x_dict to correct for reverse flux.
        for reverse in reverse_reactions:
            forward = reverse.reflection
            x_dict[forward.id] -= x_dict.pop(reverse.id)
        cobra_model.solution.x = [x_dict[r.id] for r in cobra_model.reactions]



from __future__ import print_function
#cobra.flux_analysis.moma.py: Runs the minimization of metabolic
#adjustment method described in Segre et al 2002 PNAS 99(23): 15112-7
from os import name as __name
from sys import modules as __modules
from warnings import warn
if __name == 'java':
    raise Exception("%s is not yet supported on jython"%__modules[__name__])
from copy import deepcopy
from time import time
from math import ceil, floor

#The next four imports need to be dealt with to obtain jython compatibilty
from numpy import array, hstack, vstack, matrix, sum
from scipy.sparse import eye, lil_matrix, dok_matrix
from scipy.sparse import hstack as s_hstack
from scipy.sparse import vstack as s_vstack

from ..core import Reaction, Metabolite
from ..manipulation import initialize_growth_medium, delete_model_genes
from ..manipulation.modify import convert_to_irreversible
from ..external.six import iteritems

from warnings import warn

def moma(wt_model, mutant_model, objective_sense='maximize', solver=None,
         tolerance_optimality=1e-8, tolerance_feasibility=1e-8,
         minimize_norm=False, the_problem='return', lp_method=0,
         combined_model=None, norm_type='euclidean'):
    """Runs the minimization of metabolic adjustment method described in
    Segre et al 2002 PNAS 99(23): 15112-7.

    wt_model: A cobra.Model object

    mutant_model: A cobra.Model object with different reaction bounds vs wt_model.
    To simulate deletions

    objective_sense: 'maximize' or 'minimize'

    solver: 'gurobi', 'cplex', or 'glpk'.  Note: glpk cannot be used with
    norm_type 'euclidean'

    tolerance_optimality: Solver tolerance for optimality.

    tolerance_feasibility: Solver tolerance for feasibility.

    the_problem: None or a problem object for the specific solver that can be
    used to hot start the next solution.

    lp_method: The method to use for solving the problem.  Depends on the solver.  See
    the cobra.flux_analysis.solvers.py file for more info.
        For norm_type == 'euclidean':
            the primal simplex works best for the test model (gurobi: lp_method=0, cplex: lp_method=1)
    
    combined_model: an output from moma that represents the combined optimization
    to be solved.  when this is not none.  only assume that bounds have changed
    for the mutant or wild-type.  This saves 0.2 seconds in stacking matrices.


    NOTE: Current function makes too many assumptions about the structures of the models


    """
    if solver is None:
        if norm_type == "euclidiean":
            solver = get_solver_name(qp=True)
        else:
            solver = get_solver_name()  # linear is not even implemented yet
    if combined_model is not None or the_problem not in ['return']:
        warn("moma currently does not support reusing models or problems. " +\
             "continuing without them")
        combined_model = None
        the_problem = 'return'
    if solver.lower() == 'cplex' and lp_method == 0:
        #print 'for moma, solver method 0 is very slow for cplex. changing to method 1'
        lp_method = 1
    if solver.lower() == 'glpk' and norm_type == 'euclidean':
        try:
            from gurobipy import Model
            solver = 'gurobi'
            warn("GLPK can't solve quadratic problems like MOMA.  Switched solver to %s"%solver)
        except:
            warn("GLPK can't solve quadratic problems like MOMA.  Switching to linear MOMA")

    if norm_type == 'euclidean':
        #Reusing the basis can get the solver stuck.
        reuse_basis = False
    if combined_model and combined_model.norm_type != norm_type:
        print('Cannot use combined_model.norm_type = %s with user-specified norm type'%(combined_model.norm_type,
                                                                                        norm_type))
        print('Defaulting to user-specified norm_type')
        combined_model = None



    number_of_reactions_in_common = len(set([x.id for x in wt_model.reactions]).intersection([x.id for x in mutant_model.reactions]))
    number_of_reactions = len(wt_model.reactions) + len(mutant_model.reactions)

    #Get the optimal wt objective value and adjust based on optimality tolerances
    wt_model.optimize(solver=solver)
    wt_optimal = deepcopy(wt_model.solution.f)
    if objective_sense == 'maximize':
        wt_optimal = floor(wt_optimal/tolerance_optimality)*tolerance_optimality
    else:
        wt_optimal = ceil(wt_optimal/tolerance_optimality)*tolerance_optimality
    if norm_type == 'linear':
        raise Exception('linear MOMA is not currently implmented')
        quadratic_component = None
    if minimize_norm:
        raise Exception('minimize_norm is not currently implemented')
        #just worry about the flux distribution and not the objective from the wt
        combined_model = mutant_model.copy()
        #implement this: combined_model.reactions[:].objective_coefficients = -wt_solution.x_dict
    else:
        #Construct a problem that attempts to maximize the objective in the WT model while
        #solving the quadratic problem.  This new problem is constructed to try to find
        #a solution for the WT model that lies close to the mutant model.  There are
        #often multiple equivalent solutions with M matrices and the one returned
        #by a simple cobra_model.optimize call may be too far from the mutant.
        #This only needs to be adjusted if we update mutant_model._S after deleting reactions
        if not combined_model:
            #Collect the set of wt reactions contributing to the objective.
            objective_reaction_coefficient_dict = dict([(x.id, x.objective_coefficient)
                                                        for x in wt_model.reactions
                                                        if x.objective_coefficient])
            
            
            combined_model = construct_difference_model(wt_model, mutant_model, norm_type)
            #Add in the virtual objective metabolite to constrain the wt_model to the space where
            #the objective was maximal
            objective_metabolite = Metabolite('wt_optimal')
            objective_metabolite._bound = wt_optimal
            if objective_sense == 'maximize':
                objective_metabolite._constraint_sense = 'G'
            else:
                objective_metabolite._constraint_sense = 'L'

            #TODO: this couples the wt_model objective reaction to the virtual metabolite
            #Currently, assumes a single objective reaction; however, this may be extended
            [combined_model.reactions.get_by_id(k).add_metabolites({objective_metabolite: v})
             for k, v in objective_reaction_coefficient_dict.items()]
                
            
            



        if norm_type == 'euclidean':
            #Makes assumptions about the structure of combined model
            quadratic_component = s_vstack((lil_matrix((number_of_reactions, number_of_reactions + number_of_reactions_in_common )),
                                            s_hstack((lil_matrix((number_of_reactions_in_common, number_of_reactions)),
                                                      eye(number_of_reactions_in_common,number_of_reactions_in_common)))))
    
        elif norm_type == 'linear':
            quadratic_component = None

    combined_model.norm_type = norm_type
    cobra_model = combined_model

    the_problem = combined_model.optimize(objective_sense='minimize',
                                         quadratic_component=quadratic_component,
                                         solver=solver,
                                         tolerance_optimality=tolerance_optimality,
                                         tolerance_feasibility=tolerance_feasibility,
                                         lp_method=lp_method, reuse_basis=reuse_basis)

    if combined_model.solution.status != 'optimal':
        warn('optimal moma solution not found: solver status %s'%combined_model.solution.status +\
             ' returning the problem, the_combined model, and the quadratic component for trouble shooting')
        return(the_problem, combined_model, quadratic_component)
             
    solution = combined_model.solution
    mutant_dict = {}
    #Might be faster to quey based on mutant_model.reactions with the 'mutant_' prefix added
    _reaction_list = [x for x in combined_model.reactions if x.id.startswith('mutant_')]
    mutant_f = sum([mutant_model.reactions.get_by_id(x.id[len('mutant_'):]).objective_coefficient *
                    x.x for x in _reaction_list])
    mutant_dict['objective_value'] = mutant_f
    wild_type_flux_total = sum([abs(solution.x_dict[x.id]) for x in wt_model.reactions])
    mutant_flux_total = sum(abs(x.x) for x in _reaction_list)
    #Need to use the new solution as there are multiple ways to achieve an optimal solution in
    #simulations with M matrices.
    mutant_dict['status'] = solution.status
    #TODO: Deal with maximize / minimize issues for a reversible model that's been converted to irreversible
    mutant_dict['flux_difference'] = flux_difference = sum([(solution.x_dict[x.id[len('mutant_'):]]
                                                             - x.x)**2 for x in _reaction_list])
    mutant_dict['the_problem'] = the_problem
    mutant_dict['combined_model'] = combined_model
    
    del wt_model, mutant_model, quadratic_component, solution
    return(mutant_dict)


def construct_difference_model(model_1, model_2, norm_type='euclidean'):
    """Combine two models into a larger model that is designed to calculate differences
    between the models

    """
    #Get index mappings
    common_dict = {}
    #Using copies of the models so things are modified above
    combined_model = model_1 = model_1.copy()
    model_2 = model_2.copy()
    for reaction_1 in model_1.reactions:
        try:
            reaction_2 = model_2.reactions.get_by_id(reaction_1.id)
            common_dict[reaction_1] = reaction_2
        except:
            continue
            
    #Add a prefix in front of the mutant_model metabolites and reactions to prevent
    #name collisions in DictList
    for the_dict_list in [model_2.metabolites,
                          model_2.reactions]:
        [setattr(x, 'id', 'mutant_%s'%x.id)
         for x in the_dict_list]
        the_dict_list._generate_index() #Update the DictList.dicts

    
    combined_model.add_reactions(model_2.reactions)
    [setattr(x, 'objective_coefficient', 0.)
     for x in combined_model.reactions]
    #Add in the difference reactions.  The mutant reactions and metabolites are already added.
    #This must be a list to maintain the correct order when adding the difference_metabolites
    difference_reactions = [] #Add the difference reactions at the end to speed things up
    difference_metabolites = []
    for reaction_1, reaction_2 in iteritems(common_dict):
        reaction_1._difference_partner = reaction_2
        reaction_2._difference_partner = reaction_1
        difference_reaction = Reaction('difference_%s'%reaction_1.id)
        difference_reactions.append(difference_reaction)
        difference_reaction.upper_bound = 100000
        difference_reaction.lower_bound = -1* difference_reaction.upper_bound
        difference_metabolite = Metabolite('difference_%s'%reaction_1.id)
        difference_metabolites.append(difference_metabolite)
        if norm_type == 'linear':
            difference_metabolite._constraint_sense = 'G'
        reaction_1.add_metabolites({difference_metabolite: -1.}, add_to_container_model=False)
        reaction_2.add_metabolites({difference_metabolite: 1.}, add_to_container_model=False)
        difference_reaction.add_metabolites({difference_metabolite: 1.}, add_to_container_model=False)

    combined_model.add_metabolites(difference_metabolites)
    combined_model.add_reactions(difference_reactions)
    return(combined_model)




from unittest import TestCase, TestLoader, TextTestRunner, skipIf

from warnings import warn
import sys
from os import name

try:
    import numpy
except:
    numpy = None

if __name__ == "__main__":
    sys.path.insert(0, "../..")
    from cobra.test import create_test_model
    from cobra import Model, Reaction, Metabolite
    from cobra.manipulation import initialize_growth_medium
    from cobra.solvers import solver_dict, get_solver_name
    from cobra.manipulation import modify, delete
    from cobra.flux_analysis.parsimonious import optimize_minimal_flux
    from cobra.flux_analysis.variability import flux_variability_analysis
    from cobra.flux_analysis.variability import find_blocked_reactions
    from cobra.flux_analysis.single_deletion import single_deletion
    from cobra.flux_analysis.loopless import construct_loopless_model
    from cobra.external.six import iteritems
    if numpy:
        from cobra.flux_analysis.double_deletion import double_deletion
    sys.path.pop(0)
else:
    from . import create_test_model
    from .. import Model, Reaction, Metabolite
    from ..manipulation import initialize_growth_medium
    from ..solvers import solver_dict, get_solver_name
    from ..manipulation import modify, delete
    from ..flux_analysis.parsimonious import optimize_minimal_flux
    from ..flux_analysis.variability import flux_variability_analysis
    from ..flux_analysis.variability import find_blocked_reactions
    from ..flux_analysis.single_deletion import single_deletion
    from ..flux_analysis.loopless import construct_loopless_model
    from ..external.six import iteritems
    if numpy:
        from ..flux_analysis.double_deletion import double_deletion


class TestCobraFluxAnalysis(TestCase):
    """Test the simulation functions in cobra.flux_analysis"""

    def setUp(self):
        self.model = create_test_model()

    def test_pFBA(self):
        model = self.model
        for solver in solver_dict:
            optimize_minimal_flux(model, solver=solver)
            abs_x = [abs(i) for i in model.solution.x]
            self.assertAlmostEqual(model.solution.f, 0.3800, places=3)
            self.assertAlmostEqual(sum(abs_x), 343.021, places=3)

    def test_modify_reversible(self):
        model1 = self.model
        model1.optimize()
        model2 = create_test_model()
        modify.convert_to_irreversible(model2)
        model2.optimize()
        self.assertAlmostEqual(model1.solution.f, model2.solution.f, places=3)
        modify.revert_to_reversible(model2)
        model2.optimize()
        self.assertAlmostEqual(model1.solution.f, model2.solution.f, places=3)

        # Ensure revert_to_reversible is robust to solutions generated both
        # before and after reversibility conversion, or not solved at all.
        model3 = create_test_model()
        model3.optimize()
        modify.convert_to_irreversible(model3)
        modify.revert_to_reversible(model3)
        self.assertAlmostEqual(model1.solution.f, model3.solution.f, places=3)

        model4 = create_test_model()
        modify.convert_to_irreversible(model4)
        modify.revert_to_reversible(model4)

    def test_gene_knockout_computation(self):
        cobra_model = self.model

        # helper functions for running tests
        delete_model_genes = delete.delete_model_genes
        find_gene_knockout_reactions = delete.find_gene_knockout_reactions

        def find_gene_knockout_reactions_fast(cobra_model, gene_list):
            compiled_rules = delete.get_compiled_gene_reaction_rules(
                cobra_model)
            return find_gene_knockout_reactions(
                cobra_model, gene_list,
                compiled_gene_reaction_rules=compiled_rules)

        def get_removed(m):
            return {x.id for x in m._trimmed_reactions}

        def test_computation(m, gene_ids, expected_reaction_ids):
            genes = [m.genes.get_by_id(i) for i in gene_ids]
            expected_reactions = {m.reactions.get_by_id(i)
                                  for i in expected_reaction_ids}
            removed1 = set(find_gene_knockout_reactions(m, genes))
            removed2 = set(find_gene_knockout_reactions_fast(m, genes))
            self.assertEqual(removed1, expected_reactions)
            self.assertEqual(removed2, expected_reactions)
            delete.delete_model_genes(m, gene_ids, cumulative_deletions=False)
            self.assertEqual(get_removed(m), expected_reaction_ids)
            delete.undelete_model_genes(m)

        gene_list = ['STM1067', 'STM0227']
        dependent_reactions = {'3HAD121', '3HAD160', '3HAD80', '3HAD140',
                               '3HAD180', '3HAD100', '3HAD181', '3HAD120',
                               '3HAD60', '3HAD141', '3HAD161', 'T2DECAI',
                               '3HAD40'}
        test_computation(cobra_model, gene_list, dependent_reactions)
        test_computation(cobra_model, ['STM4221'], {'PGI'})
        test_computation(cobra_model, ['STM1746.S'], {'4PEPTabcpp'})
        # test cumulative behavior
        delete_model_genes(cobra_model, gene_list[:1])
        delete_model_genes(cobra_model, gene_list[1:],
                           cumulative_deletions=True)
        delete_model_genes(cobra_model, ["STM4221"],
                           cumulative_deletions=True)
        dependent_reactions.add('PGI')
        self.assertEqual(get_removed(cobra_model), dependent_reactions)
        # non-cumulative following cumulative
        delete_model_genes(cobra_model, ["STM4221"],
                           cumulative_deletions=False)
        self.assertEqual(get_removed(cobra_model), {'PGI'})
        # make sure on reset that the bounds are correct
        reset_bound = cobra_model.reactions.get_by_id("T2DECAI").upper_bound
        self.assertEqual(reset_bound, 1000.)
        # test computation when gene name is a subset of another
        test_model = Model()
        test_reaction_1 = Reaction("test1")
        test_reaction_1.gene_reaction_rule = "eggs or (spam and eggspam)"
        test_model.add_reaction(test_reaction_1)
        test_computation(test_model, ["eggs"], set())
        test_computation(test_model, ["eggs", "spam"], {'test1'})
        # test computation with nested boolean expression
        test_reaction_1.gene_reaction_rule = \
            "g1 and g2 and (g3 or g4 or (g5 and g6))"
        test_computation(test_model, ["g3"], set())
        test_computation(test_model, ["g1"], {'test1'})
        test_computation(test_model, ["g5"], set())
        test_computation(test_model, ["g3", "g4", "g5"], {'test1'})
        # test computation when gene names are python expressions
        test_reaction_1.gene_reaction_rule = "g1 and (for or in)"
        test_computation(test_model, ["for", "in"], {'test1'})
        test_computation(test_model, ["for"], set())
        test_reaction_1.gene_reaction_rule = "g1 and g2 and g2.conjugate"
        test_computation(test_model, ["g2"], {"test1"})
        test_computation(test_model, ["g2.conjugate"], {"test1"})
        test_reaction_1.gene_reaction_rule = "g1 and (try:' or 'except:1)"
        test_computation(test_model, ["try:'"], set())
        test_computation(test_model, ["try:'", "'except:1"], {"test1"})

    def test_single_deletion(self):
        cobra_model = self.model
        initialize_growth_medium(cobra_model, 'LB')

        # Expected growth rates for the salmonella model with deletions in LB
        the_loci = ['STM4081', 'STM0247', 'STM3867', 'STM2952']
        the_genes = tpiA, metN, atpA, eno = map(cobra_model.genes.get_by_id,
                                                the_loci)
        id_to_name = dict([(x.id, x.name) for x in the_genes])
        growth_dict = {'fba': {tpiA.id: 2.41, metN.id: 2.44,
                               atpA.id: 1.87, eno.id: 1.81},
                       'moma': {tpiA.id: 1.62, metN.id: 2.4,
                                atpA.id: 1.40, eno.id: 0.33}}

        # MOMA requires cplex or gurobi
        try:
            get_solver_name(qp=True)
        except:
            growth_dict.pop('moma')
        for method, the_growth_rates in growth_dict.items():
            element_list = the_growth_rates.keys()
            results = single_deletion(cobra_model, element_list=element_list,
                                      method=method)
            rates = results[0]
            statuses = results[1]

            for the_gene in element_list:
                self.assertEqual(statuses[the_gene], 'optimal')
                self.assertAlmostEqual(rates[the_gene],
                                       the_growth_rates[the_gene],
                                       places=2)

    def test_single_reaction_deletion(self):
        cobra_model = self.model
        results, status = single_deletion(
            cobra_model, element_list=cobra_model.reactions[100:110:2],
            element_type="reaction")
        self.assertEqual(len(results), 5)
        self.assertEqual(len(status), 5)
        for status_value in status.values():
            self.assertEqual(status_value, "optimal")
        expected_results = {'3OAS140': 0, '3OAS160': 0.38001,
                            '3OAS180': 0.38001, '3OAS60': 0,
                            '3PEPTabcpp': 0.38001}
        for reaction, value in results.items():
            self.assertAlmostEqual(value, expected_results[reaction], 5)

    @skipIf(numpy is None, "double deletions require numpy")
    def test_double_deletion(self):
        cobra_model = self.model
        # turn into a double deletion unit test
        initialize_growth_medium(cobra_model, 'LB')
        # Expected growth rates for the salmonella model with deletions in LB
        the_loci = ['STM4081', 'STM0247', 'STM3867', 'STM2952']
        the_genes = list(map(cobra_model.genes.get_by_id, the_loci))
        growth_dict = {}
        growth_list = [[2.41, 2.389, 1.775, 1.81],
                       [2.389, 2.437, 1.86, 1.79],
                       [1.775, 1.86, 1.87, 1.3269],
                       [1.81, 1.79, 1.3269, 1.81]]
        for the_gene, the_rates in zip(the_genes, growth_list):
            growth_dict[the_gene] = dict(zip(the_genes, the_rates))

        the_solution = double_deletion(cobra_model, element_list_1=the_genes,
                                       element_list_2=the_genes)
        # Potential problem if the data object doesn't have a tolist function
        s_data = the_solution['data'].tolist()
        s_x = the_solution['x']
        s_y = the_solution['y']
        for gene_x, rates_x in zip(s_x, s_data):
            for gene_y, the_rate in zip(s_y, rates_x):
                self.assertAlmostEqual(
                    growth_dict[gene_x][gene_y], the_rate, places=2,
                    msg="%.3f != %.3f for (%s, %s)" %
                    (growth_dict[gene_x][gene_y], the_rate,
                     gene_x.id, gene_y.id))

    def test_flux_variability(self):
        fva_results = {
            '5DGLCNtex': {'minimum': 0.0, 'maximum': 0.0},
            'ABTA': {'minimum': 0.0, 'maximum': 0.0},
            '5DOAN': {'minimum': 0.0, 'maximum': 0.0},
            'A5PISO': {'minimum': 0.00692, 'maximum': 0.00692},
            'AACPS1': {'minimum': 0.0, 'maximum': 0.0},
            'AACPS2': {'minimum': 0.0, 'maximum': 0.0},
            'ACALDtex': {'minimum': 0.0, 'maximum': 0.0},
            'AACPS3': {'minimum': 0.0, 'maximum': 0.0},
            'AACPS4': {'minimum': 0.0, 'maximum': 0.0},
            'ABUTD': {'minimum': 0.0, 'maximum': 0.0},
            'AACPS5': {'minimum': 0.0, 'maximum': 0.0},
            'AACPS6': {'minimum': 0.0, 'maximum': 0.0},
            'AACPS7': {'minimum': 0.0, 'maximum': 0.0},
            'AACPS8': {'minimum': 0.0, 'maximum': 0.0},
            'AACPS9': {'minimum': 0.0, 'maximum': 0.0},
            'AACTOOR': {'minimum': 0.0, 'maximum': 0.0},
            'ABUTt2pp': {'minimum': 0.0, 'maximum': 0.0},
            '3OAS140': {'minimum': 0.50419, 'maximum': 0.50419},
            '3OAS141': {'minimum': 0.03748, 'maximum': 0.03748},
            '3OAS160': {'minimum': 0.41769, 'maximum': 0.41769},
            '3OAS161': {'minimum': 0.03748, 'maximum': 0.03748},
            '3OAS180': {'minimum': 0.01071, 'maximum': 0.01071},
            '3OAS181': {'minimum': 0.01606, 'maximum': 0.01606},
            'ABUTtex': {'minimum': 0.0, 'maximum': 0.0},
            '3OAS60': {'minimum': 0.54399, 'maximum': 0.54399},
            '3OAS80': {'minimum': 0.54399, 'maximum': 0.54399},
            'AAMYL': {'minimum': 0.0, 'maximum': 0.0},
            '3PEPTabcpp': {'minimum': 0.0, 'maximum': 0.0},
            '3PEPTtex': {'minimum': 0.0, 'maximum': 0.0},
            '3UMPtex': {'minimum': 0.0, 'maximum': 0.0},
            '4HOXPACDtex': {'minimum': 0.0, 'maximum': 0.0},
            'ACACtex': {'minimum': 0.0, 'maximum': 0.0},
            '4PCP': {'minimum': 0.0, 'maximum': 0.0},
            '4PCPpp': {'minimum': 0.0, 'maximum': 0.0},
            'AAMYLpp': {'minimum': 0.0, 'maximum': 0.0},
            '4PEPTabcpp': {'minimum': 0.0, 'maximum': 0.0},
            '4PEPTtex': {'minimum': 0.0, 'maximum': 0.0},
            '5DGLCNR': {'minimum': 0.0, 'maximum': 0.0},
            '5DGLCNt2rpp': {'minimum': 0.0, 'maximum': 0.0},
            'ACALD': {'minimum': 3.35702, 'maximum': 7.49572}}

        infeasible_model = create_test_model()
        infeasible_model.reactions.get_by_id("EX_glyc_e").lower_bound = 0
        for solver in solver_dict:
            cobra_model = create_test_model()
            initialize_growth_medium(cobra_model, 'LB')
            fva_out = flux_variability_analysis(
                cobra_model, solver=solver,
                reaction_list=cobra_model.reactions[100:140:2])
            for the_reaction, the_range in iteritems(fva_out):
                for k, v in iteritems(the_range):
                    self.assertAlmostEqual(fva_results[the_reaction][k], v,
                                           places=5)
            # ensure that an infeasible model does not run FVA
            self.assertRaises(ValueError, flux_variability_analysis,
                              infeasible_model, solver=solver)

    def test_find_blocked_reactions(self):
        m = self.model
        result = find_blocked_reactions(m, reaction_list=m.reactions[:10])
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], '12PPDStex')

    def test_loopless(self):
        try:
            solver = get_solver_name(mip=True)
        except:
            self.skip("no MILP solver found")
        test_model = Model()
        test_model.add_metabolites(Metabolite("A"))
        test_model.add_metabolites(Metabolite("B"))
        test_model.add_metabolites(Metabolite("C"))
        EX_A = Reaction("EX_A")
        EX_A.add_metabolites({test_model.metabolites.A: 1})
        DM_C = Reaction("DM_C")
        DM_C.add_metabolites({test_model.metabolites.C: -1})
        v1 = Reaction("v1")
        v1.add_metabolites({test_model.metabolites.A: -1,
                            test_model.metabolites.B: 1})
        v2 = Reaction("v2")
        v2.add_metabolites({test_model.metabolites.B: -1,
                            test_model.metabolites.C: 1})
        v3 = Reaction("v3")
        v3.add_metabolites({test_model.metabolites.C: -1,
                            test_model.metabolites.A: 1})
        DM_C.objective_coefficient = 1
        test_model.add_reactions([EX_A, DM_C, v1, v2, v3])
        feasible_sol = construct_loopless_model(test_model).optimize()
        v3.lower_bound = 1
        infeasible_sol = construct_loopless_model(test_model).optimize()
        self.assertEqual(feasible_sol.status, "optimal")
        self.assertEqual(infeasible_sol.status, "infeasible")

# make a test suite to run all of the tests
loader = TestLoader()
suite = loader.loadTestsFromModule(sys.modules[__name__])


def test_all():
    TextTestRunner(verbosity=2).run(suite)

if __name__ == "__main__":
    test_all()
